/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
Developer: Robert Tupelo-Schneck <schneck@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.apps.doserver;

import net.cnri.apps.doserver.txnlog.*;
import net.cnri.apps.doserver.operators.*;
import net.cnri.dobj.*;
import net.cnri.dobj.delegation.DelegationClient;
import net.cnri.knowbots.station.*;
import net.cnri.simplexml.*;
import net.cnri.util.IOForwarder;
import net.cnri.util.StreamTable;
import net.cnri.util.StreamVector;
import net.cnri.util.StringUtils;
import net.cnri.util.ThreadSafeDateFormat;
import net.cnri.util.handlerpool.*;
import net.handle.hdllib.*;

import java.io.*;
import java.net.*;
import javax.net.ssl.*;

import org.eclipse.jetty.deploy.App;
import org.eclipse.jetty.deploy.DeploymentManager;
import org.eclipse.jetty.deploy.providers.WebAppProvider;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.ContextHandler;
import org.eclipse.jetty.server.handler.ContextHandlerCollection;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.server.ssl.SslSelectChannelConnector;
import org.eclipse.jetty.servlet.*;
import org.eclipse.jetty.util.ssl.SslContextFactory;
import org.eclipse.jetty.webapp.WebAppContext;

import java.security.*;
import java.security.cert.*;
import java.util.*;

/**
 * This class runs a server object that accepts and handles incoming requests
 * for operations on digital objects.  Agents in an embedded knowbot service 
 * station are used to determine access control as well as to perform operations
 * on objects in the server.
 */
public class Main {
    public static final boolean DEBUG = false;

    public static final String SERVER_INFO_FILE = "server_info.xml";
    public static final String PRIVATE_KEY_FILE = "privatekey";
    public static final String PUBLIC_KEY_FILE = "publickey";
    public static final String STORED_PASSWORDS_FILE = "password.dct";
    public static final int PUBLIC_KEY_HDL_IDX = 300;
    public static final int DEFAULT_PORT = 9900;
    public static final int DEFAULT_HTTP_PORT = 0; //8800;
    public static final int DEFAULT_SSL_PORT = 0;
    public static final int DEFAULT_HTTPS_PORT = 0; // 8443
    public static final int DOP_LISTEN_BACKLOG = 5;
    public static final int DOPSSL_LISTEN_BACKLOG = 5;
    
	private static final int NUM_PRELOADED_CONNECTION_HANDLERS = 2;
	private static final int NUM_PRELOADED_OP_HANDLERS = 4;
	private static final int NUM_CONNECTIONS_PER_HANDLER = 50;
	private static final int NUM_OPS_PER_HANDLER = 500;

	private static final boolean DEFAULT_REGISTER_IDS = false;
	private static final boolean DEFAULT_ENABLE_AUDIT_LOG = false;
	private static final boolean DEFAULT_DO_PUSH_REPLICATION = false;
  private static final boolean DEFAULT_DO_PULL_REPLICATION = true;
  
  public static final String OWNER_ATTRIBUTES_KEY = "owner_attributes";
  public static final String AUTH_DELEGATES_KEY = "authorization_delegate_attributes";
  public static final String DO_PUSH_REPLICATION_KEY = "enable_push_replication";
  public static final String DO_PULL_REPLICATION_KEY = "enable_pull_replication";
	public static final String ENABLE_AUDIT_LOGS_KEY = "enable_audit_logs";
	public static final String REGISTER_HANDLES_KEY = "register_handles";
	public static final String AUTO_OBJECT_PREFIX_KEY = "handle_prefix";
  public static final String SERVER_ADMIN_KEY = "server_admin";
	public static final String PORT_KEY = "port";
	public static final String BIND_KEY = "listen_address";
	public static final String ADDRESS_KEY = "public_address";
	public static final String REDIRECT_STDERR_KEY = "redirect_stderr";
    public static final String LOG_ACCESSES_KEY = "log_accesses";
    public static final String LOG_SAVE_INTERVAL_KEY = "log_save_interval";
    public static final String MONTHLY = "Monthly";
    public static final String WEEKLY = "Weekly";
    public static final String DAILY = "Daily";
    public static final String NEVER = "Never";
	public static final String HTTP_PORT_KEY = "http_port";
	public static final String SSL_PORT_KEY = "ssl_port";
  public static final String HTTPS_PORT_KEY = "https_port";
  public static final String HTTPS_HOSTNAME_KEY = "https_hostname";
	public static final String STORAGE_CLASS_KEY = "storage_class";
	public static final String STORAGE_DIR_KEY = "storage_directory";
	public static final String SERVER_ID_KEY = "serverid";
	public static final String SERVICE_ID_KEY = "serviceid";
  public static final String HTTPS_KEY_STORE_FILE = "https.key";
  public static final String LISTEN_IP_KEY = "listen_addr";
  public static final String AUTO_SET_PERMISSIONS_FOR_NEW_OBJECTS = "auto_set_permissions";
  private static final String KNOWBOT_ELEMENT_PREFIX = "internal.preload_knowbot_";
	public static final String VERSION_ATT = "internal.version";
	private static SetupHelper setup = null;
  
	private Main thisServer = null;
	private boolean auditLogEnabled = DEFAULT_ENABLE_AUDIT_LOG;
	private PKAuthentication myID = null;
  private PublicKey myPublicKey = null;
	private PublicKeyAuthenticationInfo hdlID = null;
	private Storage storage = null;
	private ExecutionStation knowbotStation = null;
	private Configuration configuration = null;
	private AuditLog auditLog = null;
	private Authorizer authorizer = null;
	private File baseDir;
	private String serverID = null;
  private String serverAdmin = null;
  private String localServerID = null;
  private String[] ownerAttributes = null; // attribute keys that specify an owner/administrator of the DO
  private String[] authorizationDelegateAttributes; // attribute keys that specify an object that grants additional rights
  private DOClient doClient = null;
	private DelegationClient delegationClient = null;
	private ReplicationManager replicator = null;
	private StreamTable serverConfig = new StreamTable();
	private final StreamTable storedPasswords = new StreamTable();
	private StreamTable nextIDConfig = new StreamTable();
	private File nextIDFile = null;
	private ServerLog logger = null;
	private ServerLog errLogger = null;
	private ThreadSafeDateFormat dateFmt = null;
  private InetAddress listenAddress = null;
	private SSLServerSocketFactory sslSocketFactory = null;
	private ServerSocket socket = null;
	private ServerSocket httpSocket = null;
	private SSLServerSocket sslSocket = null;

	private boolean keepServing = false;

	private RequestHandlerPool connPool = null;
	private RequestHandlerPool opHandlerPool = null;

  private static final void printUsageAndExit() {
		System.err.println("usage: java net.cnri.apps.doserver.Main <serverdir>"); 
		System.err.println("   or: java net.cnri.apps.doserver.Main -setup <serverdir> [-knowbotid=<knowbot name> -knowbotloc=<path to knowbot>]");
		System.err.println("");
		System.exit(1);
  }
  
  public static void main(String argv[])
	throws Exception
	{
	    System.out.println("DO Repository Server Software version " + Version.version);
	    
		boolean setupMode = false;
		File serverDir = null;
		String knowbotID = null;
		String knowbotLoc = null;

		//look for setup and knowbot flags
		for(int i=0; i<argv.length; i++) {
			String arg = argv[i];
			if(arg.startsWith("-")) {
				if(arg.equals("-setup")) {
					setupMode = true;
				}
			} else if(setupMode && arg.startsWith("-knowbotid=")) {
				knowbotID = StringUtils.fieldIndex(arg, '=', 1);
			} else if(setupMode && arg.startsWith("-knowbotfile=")) {
				knowbotID = StringUtils.fieldIndex(arg, '=', 1);
			} else if(serverDir==null) {
				serverDir = new File(arg);
			} else {
				printUsageAndExit();
			}
		}

		if(serverDir==null || (knowbotID==null && knowbotLoc!=null) || (knowbotID!=null && knowbotLoc==null))
			printUsageAndExit();

		//Read key files

        if(DEBUG) System.out.println("*Reading Keys");
		File privKeyFile = new File(serverDir, PRIVATE_KEY_FILE);
        File pubKeyFile = new File(serverDir, PUBLIC_KEY_FILE);
        
		//Setup, if setup flag is set
		if(setupMode) {
			setup = new SetupHelper();

			if(!serverDir.exists())
				serverDir.mkdirs();

			System.out.println("\nTo configure your new or existing object server,"
					+ "\nplease answer the questions which follow; default"
					+ "\nanswers, shown in [square brackets] when available,"
					+ "\ncan be chosen by pressing Enter.");

			boolean generateKeys = true;
			if(privKeyFile.exists() && pubKeyFile.exists()) {
				// there is already a public and private key... ask the user if we should re-use them
				generateKeys = setup.getBoolean("Keys already exist.  Generate new ones? ", false);
			}

			if(generateKeys) {
				setup.generateKeys(pubKeyFile, privKeyFile, "Server Authentication");
			}
		}

		//Check if privatekey is accessible
		if(!privKeyFile.exists() || !privKeyFile.canRead()) {
			if(DEBUG)
				System.out.println("*ERROR: Cannot read Key Files");
			System.err.println("Error:  cannot read private key file: "+privKeyFile.getAbsolutePath());
			System.exit(1);
		}

		//Setup input and output Streams
		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		FileInputStream fin = new FileInputStream(privKeyFile);

		byte buf[] = new byte[1024];
		int r;
		while((r=fin.read(buf))>=0)
			bout.write(buf, 0, r);
		fin.close();
		buf = bout.toByteArray();
		byte secKey[] = null;
		if(Util.requiresSecretKey(buf)) {
          if(DEBUG) System.err.println("\nReading the private key...");
          secKey = Util.getPassphrase("Enter the passphrase to decrypt the private key in "+
                                      privKeyFile.getAbsolutePath());
        } else {
          if(DEBUG) System.err.println("*Does not Require Secret Key");
        }
		buf = Util.decrypt(buf, secKey);
		PrivateKey svrKey = Util.getPrivateKeyFromBytes(buf, 0);

		// load the server...
        if(DEBUG) System.err.println("*Loading server");
		Main runner = new Main(svrKey, serverDir, setup);
		
		// if loading a knowbot in setup mode...
		if(setupMode && knowbotID!=null && knowbotLoc!=null) {
			File knowbotFile = new File(knowbotLoc);
			if(knowbotFile.exists() && knowbotFile.canRead()) {
				runner.addOperator(knowbotID, new FileInputStream(knowbotFile));
			}
		}

		if(setupMode) {
			System.exit(0);
		} else {
			runner.serveRequests();
		}
	}
  
	public Main(PrivateKey serverKey, File baseDir, SetupHelper setup)
	throws Exception
	{
		System.out.println();
		this.baseDir = baseDir;
		this.thisServer = this;
        this.myPublicKey = Util.getPublicKeyFromFile(new File(baseDir, PUBLIC_KEY_FILE).getAbsolutePath());

		File configFile = new File(baseDir, "config.dct");

		// read the server-specific (as opposed to site-specific) configuration
		try {
			serverConfig.clear();
			if(configFile.exists())
				serverConfig.readFromFile(configFile);
		} catch (Exception e) {
			System.err.println("Error loadig config file: "+e);
			throw e;
		}
		if(DEBUG)
			System.out.println("Read from Config File");

        // read the stored passwords
        try {
            File storedPasswordFile = new File(baseDir, STORED_PASSWORDS_FILE);
            storedPasswords.clear();
            if(storedPasswordFile.exists())
                storedPasswords.readFromFile(storedPasswordFile);
        } catch (Exception e) {
            System.err.println("Error loadig stored passwords: "+e);
        }

        nextIDFile = new File(baseDir, "next_id.dct");
		nextIDConfig.clear();
		if(nextIDFile.exists()) {
			nextIDConfig.readFromFile(nextIDFile);
		}

		this.serverID = serverConfig.getStr(SERVICE_ID_KEY, null);

		this.localServerID = serverConfig.getStr(SERVER_ID_KEY, "");
		this.dateFmt = new ThreadSafeDateFormat("yyyyMMdd HH:mm:ss:SSS");
        
        if(setup!=null) {
			String description = setup.getServerDescription();
			InetAddress externalAddr = setup.getExternalAddress();
			InetAddress internalAddr = setup.getInternalAddress();
			int port = setup.getListenPort();
			int sslPort = setup.getSSLPort();
			int httpPort = setup.getHTTPPort();
            int httpsPort = setup.getHTTPSPort();
            String adminStr = setup.responseToPrompt("Enter the handle for the server admin, " +
                                                     "who will be allowed to perform any operation " +
                                                     "on this server");
            if(adminStr!=null && adminStr.trim().length()<=0) adminStr = null;
            
			ByteArrayOutputStream bout = new ByteArrayOutputStream();
			FileInputStream fin = new FileInputStream(new File(baseDir, PUBLIC_KEY_FILE));
			byte buf[] = new byte[1024];
			int r;
			while((r=fin.read(buf))>=0) bout.write(buf, 0, r);
			byte pubKeyBytes[] = bout.toByteArray();

			while(true) {
//			    boolean hosted = setup.getHosted();
                boolean hosted = false;
			    if(hosted) {
			        serverID = null;
			        localServerID = "1";
			    } else {
			        serverID = setup.getUniqueServiceID();
			        localServerID = setup.getUniqueServerID();
			        if(!setup.verifyID(serverID, localServerID)) continue;
			    }
			        
				serverID = setup.registerServer(serverID, localServerID, pubKeyBytes, externalAddr,
						port, sslPort, httpPort, httpsPort, description, new File(baseDir, "serverinfo.xml"), 
						new File(baseDir, PUBLIC_KEY_FILE), new File(baseDir, PRIVATE_KEY_FILE), serverKey);
				break;
			}
			serverConfig.put(REDIRECT_STDERR_KEY, setup.getRedirectStdErr());
            serverConfig.put(LOG_ACCESSES_KEY, setup.getLogAccesses());
            serverConfig.put(LOG_SAVE_INTERVAL_KEY, setup.getLogSaveInterval());
			serverConfig.put(SERVICE_ID_KEY, serverID);
			serverConfig.put(SERVER_ID_KEY, localServerID);
			serverConfig.put(PORT_KEY, port);
			serverConfig.put(HTTP_PORT_KEY, httpPort);
            serverConfig.put(HTTPS_PORT_KEY, httpsPort);
			serverConfig.put(SSL_PORT_KEY, sslPort);
			serverConfig.put(BIND_KEY, internalAddr.getHostAddress());
			serverConfig.put(ADDRESS_KEY, externalAddr.getHostAddress());
			serverConfig.put(AUTO_OBJECT_PREFIX_KEY, setup.getNewObjectPrefix(serverID, localServerID));
			serverConfig.put(REGISTER_HANDLES_KEY, false);
            serverConfig.put(SERVER_ADMIN_KEY, adminStr);  
			try {
				serverConfig.writeToFile(configFile);
			} catch (Exception e) {
				System.err.println("Error saving config file ("+configFile.getAbsolutePath()+"): "+e);
				throw e;
			}
			// end registration of new server ID
			// build the default configuration for the server
			XTag newConfig = new XTag("REPOCONFIG");
			XTag knowbotList = new XTag("KNOWBOTS");
			newConfig.addSubTag(knowbotList);
		} // end entry of server configuration

		// initialize the storage and create the repository object if it doesn't already exist
		//this.storage = new HashDirectoryStorage(this, new File(baseDir, "storage"));
    this.serverAdmin = serverConfig.getStr(SERVER_ADMIN_KEY, null);
    
    
    // get the list of attributes that identify an owner/administrator for the object
    StreamVector ownerAtts = (StreamVector)serverConfig.get(OWNER_ATTRIBUTES_KEY);
    if(ownerAtts!=null) {
      ownerAttributes = new String[ownerAtts.size()];
      int ownerIdx = 0;
      for(Object att : ownerAtts) {
        ownerAttributes[ownerIdx++] = (String)att;
      }
    } else {
      ownerAttributes = null;
    }
    
    StreamVector authDelAtts = (StreamVector)serverConfig.get(AUTH_DELEGATES_KEY);
    if(authDelAtts!=null) {
      authorizationDelegateAttributes = new String[authDelAtts.size()];
      int ownerIdx = 0;
      for(Object att : authDelAtts) {
        authorizationDelegateAttributes[ownerIdx++] = (String)att;
      }
    } else {
      authorizationDelegateAttributes = null;
    }
    
    
		String storageClass = serverConfig.getStr(STORAGE_CLASS_KEY,"net.cnri.apps.doserver.HashDirectoryStorage");
		String storageDir = serverConfig.getStr(STORAGE_DIR_KEY,"storage");
		this.storage = (Storage) Class.forName(storageClass).newInstance();
		this.storage.initWithDirectory(this, new File(baseDir,storageDir));
        
		File txnDir = new File(baseDir, "txns");
		this.storage.initTransactionQueue(txnDir);

		if(!storage.doesObjectExist(serverID)) {
			storage.createObject(serverID, null, true);
		}
    
    DOConnection.setResolver(new DOStorageResolver(DOConnection.getResolver(), thisServer));
    
		// initialize our authentication which is used for communicating with clients and
		// other servers
		this.myID = new PKAuthentication(serverID, serverKey);
		this.myID.setAutoRetrieveCredentials(false);

		// set version attribute
		HeaderSet atts = storage.getAttributes(serverID,null,null);
		String version = atts.getStringHeader(VERSION_ATT,null);
		if(!Version.version.equals(version)) {
		    atts = new HeaderSet();
		    atts.addHeader(VERSION_ATT,Version.version);
		    storage.setAttributes(serverID,null,atts,false,0);
		}
		
		// load my credentials directly from storage...
        StorageProxy myStorage = new ConcreteStorageProxy(storage, serverID, serverID, null);
		InputStream credIn = myStorage.getDataElement(StandardOperations.OBJ_CREDENTIALS_ELEMENT);

		if(credIn!=null) {
			XParser xmlParser = new XParser();
			XTag xmlCertList = xmlParser.parse(new InputStreamReader(credIn, "UTF8"),
					false);
			ArrayList certs = new ArrayList();
			for(int i=0; i<xmlCertList.getSubTagCount(); i++) {
				XTag subtag = xmlCertList.getSubTag(i);
				if(subtag.getName().equalsIgnoreCase("certificate")) {
					X509Certificate cert = null;            
					try {
						InputStream certIn =
							new ByteArrayInputStream(Util.encodeHexString(subtag.getStrSubTag("x509", "")));
						cert = (X509Certificate)StandardOperations.getCertFactory().generateCertificate(certIn);
						cert.checkValidity();
						certs.add(cert);
					} catch (Exception e) {
						e.printStackTrace(System.err);
						continue;
					}
				}
			}
			this.myID.setCredentials((java.security.cert.Certificate[])certs.toArray(new java.security.cert.Certificate[certs.size()]));
		}      

		this.hdlID = new PublicKeyAuthenticationInfo(Util.encodeString(serverID),
				PUBLIC_KEY_HDL_IDX, serverKey);
		// initialize the knowbots that handle operations
		File knowbotDir = new File(baseDir, "knowbots");
		if(setup!=null || !knowbotDir.exists()) {
			if(!knowbotDir.exists()) {
				knowbotDir.mkdirs();
			}
			StreamTable kbConfig = new StreamTable();
			kbConfig.put("station_uri", "hdl:"+serverID);
			StreamVector trustedIDs = new StreamVector();
			trustedIDs.addElement("hdl:"+serverID);
			kbConfig.put("trusted_uris", trustedIDs);
			kbConfig.writeToFile(new File(knowbotDir, "config.dct"));
			// reset the permissions
			new File(knowbotDir, "permissions").delete();
		}

		this.knowbotStation = new ExecutionStation(knowbotDir, "hdl:"+serverID, serverKey);
		sslSocketFactory = DOServerConnection.getSecureServerSocketFactory(this.myID, this.myPublicKey);

		// load up the loggers!
		File logDir = new File(baseDir, "logs");
		this.auditLog = new AuditLog(this, logDir);
		String logSaveIntervalStr = serverConfig.getStr(LOG_SAVE_INTERVAL_KEY,MONTHLY);
		int logSaveInterval;
        if(NEVER.equalsIgnoreCase(logSaveIntervalStr)) logSaveInterval = ServerLog.ROTATE_NEVER;
        else if(DAILY.equalsIgnoreCase(logSaveIntervalStr)) logSaveInterval = ServerLog.ROTATE_DAILY;
        else if(WEEKLY.equalsIgnoreCase(logSaveIntervalStr)) logSaveInterval = ServerLog.ROTATE_WEEKLY;
        else logSaveInterval = ServerLog.ROTATE_MONTHLY;
		if(serverConfig.getBoolean(LOG_ACCESSES_KEY, true)) {
		    this.logger = new ServerLog(logDir, "access.log", false,
		            logSaveInterval);
		}
		this.errLogger = new ServerLog(logDir, "error.log", 
				serverConfig.getBoolean(REDIRECT_STDERR_KEY, true),
				logSaveInterval);
		this.errLogger.log("DO Repository Server Software version " + Version.version);

		// create a default authorizer which gets to decide who can do what to which objects
		this.authorizer = new Authorizer(this);

		if(this.localServerID!=null) {
			this.replicator = new ReplicationManager(this, storage);
		}

        this.doClient = new DOClient(this.myID);
        this.delegationClient = new DelegationClient(doClient);

		this.configuration = new Configuration(this, this.storage, serverKey);

		this.auditLogEnabled = serverConfig.getBoolean(ENABLE_AUDIT_LOGS_KEY, 
				DEFAULT_ENABLE_AUDIT_LOG);

        if(this.replicator!=null && serverConfig.getBoolean(DO_PULL_REPLICATION_KEY,
                DEFAULT_DO_PULL_REPLICATION)) {
            this.replicator.doPullReplication();
        }
		if(this.replicator!=null && serverConfig.getBoolean(DO_PUSH_REPLICATION_KEY,
				DEFAULT_DO_PUSH_REPLICATION)) {
			this.replicator.doPushReplication();
		}

		if(setup!=null) {
			// sign and add the standard knowbot operators to the object server config
			// the addOperator will confirm with the user which bundled 
			// operators should be loaded
			//InputStream opsBotIn = getClass().getResourceAsStream("/standardopsbot.jar");
			//if(opsBotIn!=null) {
			//  addOperator("standardops", opsBotIn);
			//}
		}

		// shutdown nicely on ^C
		Runtime.getRuntime().addShutdownHook(new Thread() {
			public void run() {
				try { 
					shutdown();
				}
				catch (Exception e) {
					System.err.println("Error while shutting down server...");
					e.printStackTrace();
				}
			}
		});

		new ServerMonitor().start();
	}
  

  /** Returns true if newly created objects should have all access permissions for */
  public boolean getAutoSetPermissionsForNewObjects() {
    return serverConfig.getBoolean(AUTO_SET_PERMISSIONS_FOR_NEW_OBJECTS, false);
  }

  
  /** Return delegation client that is used to verify group membership */
  public DelegationClient getDelegationClient() {
    return delegationClient;
  }

  public String[] getOwnerAttributes() {
    return ownerAttributes;
  }
  
  public String[] getAuthorizationDelegateAttributes() {
    return authorizationDelegateAttributes;
  }


  private class ServerMonitor
	extends Thread
	{
		File keepRunningFile;

		ServerMonitor() 
		throws Exception
		{
			setDaemon(true);
			setName("Server Monitor");

			keepRunningFile = new File(baseDir, "delete_this_to_stop_server");
			keepRunningFile.createNewFile();
            keepRunningFile.deleteOnExit();
		}

		public void run() {
			while(true) {
				try {
					if(!keepRunningFile.exists()) {
					    shutdown();
						System.exit(0);
					}

					Thread.sleep(1000);
				} catch (Throwable t) {}
			}
		}
  }
  
  
  /** Get an iterator over the set of server configuration keys */
  public Iterator getConfigKeys() {
    return serverConfig.keySet().iterator();
  }
  
  /** Get the value from the configuration table that is associated with the given key */
  public String getConfigVal(String configKey) {
    return serverConfig.getStr(configKey);
  }

  /** 
   * Get the value from the configuration table that is associated with the given key.
   * If no value is associated with the given key then return defaultValue 
   */
  public String getConfigVal(String configKey, String defaultValue) {
    return serverConfig.getStr(configKey, defaultValue);
  }
  
  /** Associate the given value with the given key in the configuration table */
  public void setConfigVal(String configKey, String configVal) {
		serverConfig.put(configKey, configVal);
	}

	/** Get password stored in a local file rather than the handle system */
	public String getStoredPasswordForUser(String userID) {
	    return storedPasswords.getStr(userID);
	}
	
	/** 
	 * Return the folder under which all of this server's configuration and other components
	 * are located.
	 */
	public File getBaseFolder() {
		return this.baseDir;
	}

	/** Add the given knowbot to the set of operators for this repository */
	/*
  public final void addOperatorElement(String knowbotElement,
                                       InputStream knowbotStream,
                                       Signature serverSig) 
    throws Exception
  {
    // start a thread to sign the agent and stream it to an output stream that
    // will be forwarded to the storage.
    PipedOutputStream pout = new PipedOutputStream();
    PipedInputStream pin = new PipedInputStream(pout);
    AgentSignerThread ast = new AgentSignerThread("hdl:"+serverID,
                                                  serverSig,
                                                  opsBotIn,
                                                  pout);
    ast.start();
      storage.storeDataElement(serverID, knowbotElement, pin, true, false);
  }
	 */

	public void addOperator(String knowbotID, InputStream knowbotLoc) 
	throws Exception
	{
		this.configuration.addKnowbot(KNOWBOT_ELEMENT_PREFIX+knowbotID, knowbotLoc);
	}


	/**
	 * Returns the identifier for this server
	 */
	public final String getServerID() {
		return serverID;
	}

	/** Returns a direct reference to the object storage */
	public Storage getStorage() {
		return storage;
	}

	/** Gets the next object ID in the sequence.  This object ID should not already
	 * exist. */
	synchronized public final String getNextObjectID() 
	throws Exception
	{
		String prefix = serverConfig.getStr(AUTO_OBJECT_PREFIX_KEY, null);
		if(prefix==null) prefix = getServerID()+"."+localServerID+"-";

		long nextNum = nextIDConfig.getLong("nextid", 0);
		String nextObjID = null;
		while(true) {
			nextObjID = prefix + nextNum;
			if(!storage.doesObjectExist(nextObjID)) break;
			nextNum++;
		}

		nextIDConfig.put("nextid", String.valueOf(nextNum+1));
		nextIDConfig.writeToFile(nextIDFile);
		return prefix+nextNum;
	}

	/** Returns the transaction queue for this server */
	public final AbstractTransactionQueue getTxnQueue() {
		return storage.getTransactionQueue();
	}

	/** Returns true if the given DOServerInfo object refers to this server.
	 */
	public final boolean isThisServer(String aLocalServerID)
	throws Exception
	{
		if(aLocalServerID==null) return false;
		return aLocalServerID.equals(localServerID);
	}


	/**
	 * Returns the configuration object for this server
	 */
	final Configuration getConfiguration() {
		return this.configuration;
	}


	/**
	 * Returns the Signature object that can be used to sign data that is
	 * endorsed by this repository.
	 */
	public Signature getSigner() {
		return knowbotStation.getSigner();
	}

	/** Return the knowbot/operator registry in which operators and other services are registered */
	public CollaborationHub getServiceRegistry() {
		return knowbotStation.getCollaborationHub();
	}

	/** Set the given key-value mapping in the server-level table that knowbots
	 * can use to communicate. */
	public void setKnowbotMapping(String key, Object value) {
		this.knowbotStation.getCollaborationHub().registerService(key, key, value, null);
	}

	/** Returns the DOAuthentication that identifies this server.
	 */
	public DOAuthentication getAuth() {
		return this.myID;
	}

	/** Return the Authorizer that determines which operations are allowed by whom */
	public Authorizer getAuthorizer() {
		return this.authorizer;
	}

    /** Return the server administrator (ie super-user) who can perform any operation on any object.
     *  Returns null if 
     */
    public String getServerAdmin() {
        return serverAdmin;
    }

    /** Returns the public key of this server */
    public PublicKey getPublicKey() {
        return myPublicKey;
    }
    
	/**
	 * Removes all currently running knowbots.  This is usually followed by
	 * calls to loadServiceBot(InputStream) to load replacement knowbots.
	 */
	void clearServiceBots() {
		java.util.Enumeration agents = this.knowbotStation.getCurrentAgents();
		while(agents.hasMoreElements()) {
			try {
				Object agent = agents.nextElement();
				((KnowbotWrapper)agent).stop();
			} catch(Throwable t) {
				//System.err.println("Error stopping agent: "+t);
				//t.printStackTrace(System.err);
			}
		}
		if(!builtInKnowbotsLoaded) {
			// register the built-in audit operator
			this.knowbotStation.getCollaborationHub().registerService("Audit Operator",
					"net.cnri.apps.doserver.audit",
					new AuditHookImpl(auditLog),
					null);

			// register the external interface to allow remote admin operations
			this.knowbotStation.getCollaborationHub().registerService("External Admin Interface",
					"net.cnri.apps.doserver.admin",
					new ExternalInterface(this),
					null);
			this.knowbotStation.getCollaborationHub().registerService("Standard Operations",
					"net.cnri.apps.doserver.standardops",
					new StandardOperations(this,this.delegationClient),
					null);
			this.knowbotStation.getCollaborationHub().registerService("Group Operations",
                    "net.cnri.apps.doserver.groupops",
                    new ListElementDelegationOperations(),
                    null);
			try {
				this.knowbotStation.getCollaborationHub().registerService("Lucene Indexer",
						"net.cnri.knowbots.lucene.Main",
						new LuceneIndexer(this),
						null);
			} catch (Exception e) {
				System.err.println("Error loading repository indexer: "+e);
				e.printStackTrace(System.err);
			}

			if(replicator!=null) {
				// register the replication operations for other servers to replicate this one
				DOOperation replOperator = replicator.getOperator();
				if(replOperator!=null) {
					this.knowbotStation.getCollaborationHub().registerService("Replication Hooks",
							"net.cnri.apps.doserver.replication",
							replOperator,
							null);
				}
			}
			builtInKnowbotsLoaded = true;
		}
	}
	private boolean builtInKnowbotsLoaded = false;


	/**
	 * Creates the given handle, if necessary, and sets the object server for it to
	 * this object server.
	 */
	public void registerHandle(String newHandle, String objectName)
	throws DOException
	{
		if(!serverConfig.getBoolean(REGISTER_HANDLES_KEY, DEFAULT_REGISTER_IDS))
			return;

		byte repoHandle[] = Util.encodeString(myID.getID());
		Resolver resolver = DOClient.getResolver();
		HandleValue values[] = null;
		try {
			values = resolver.resolveHandle(newHandle);
		} catch (HandleException e) {
			// avoid stupid exception being thrown in resolveHandle when a
			// handle not found message is received
			if(e.getCode()!=AbstractResponse.RC_HANDLE_NOT_FOUND)
				throw new DOException(DOException.INTERNAL_ERROR,
						"Unable to locate handle: "+e);
		}

		byte hdlBytes[] = Util.encodeString(newHandle);
		if(values==null) { // handle does not exist?
			AdminRecord adminRec = new AdminRecord(hdlID.getUserIdHandle(),
					hdlID.getUserIdIndex(),
					false, // add handle
					true, // delete handle
					false, // add NA
					false, // delete NA
					true, // read value
					true, // modify value
					true, // remove value
					true, // add value
					true, // modify admin
					true, // remove admin
					true, // add admin
					false // list handles
			); 

			HandleValue hdlValues[] = null;
			if(objectName==null) {
				hdlValues = new HandleValue[] {
						new HandleValue(100, Common.ADMIN_TYPE, null),
						new HandleValue(1, Util.encodeString(DOConstants.OBJECT_SERVER_HDL_TYPE),
								repoHandle)
				};
			} else {
				hdlValues = new HandleValue[] {
						new HandleValue(100, Common.ADMIN_TYPE, null),
						new HandleValue(1, Util.encodeString(DOConstants.OBJECT_SERVER_HDL_TYPE),
								repoHandle),
								new HandleValue(2, Util.encodeString(DOConstants.OBJECT_NAME_HDL_TYPE),
										Util.encodeString(objectName))
				};
			}
			hdlValues[0].setData(Encoder.encodeAdminRecord(adminRec));
			CreateHandleRequest createRq = new CreateHandleRequest(hdlBytes,
					hdlValues,
					hdlID);
			try {
				AbstractResponse rs = resolver.getResolver().processRequest(createRq);
				if(rs!=null && rs.responseCode==AbstractMessage.RC_SUCCESS) {
					return;
				} else {
					throw new DOException(DOException.INTERNAL_ERROR,
							"Error registering handle: "+rs);
				}
			} catch (HandleException e) {
				throw new DOException(DOException.INTERNAL_ERROR,
						"Unable to register handle: "+e);
			}
		} else { // handle exists, modify if necessary/possible

			byte objSvrType[] = Util.encodeString(DOConstants.OBJECT_SERVER_HDL_TYPE);
			// check to see if it already points to us...
			for(int i=0; i<values.length; i++) {
				if(values[i]!=null && values[i].hasType(objSvrType)) {
					if(Util.decodeString(values[i].getData()).equalsIgnoreCase(myID.getID())) {
						// we are already listed as a registry server for this object
						return;
					}
				}
			}

			int newIndex = 0;
			boolean duplicateIndex = true;
			while(duplicateIndex) {
				newIndex++;
				boolean foundDupe = false;
				for(int i=0; i<values.length; i++) {
					if(values[i].getIndex()==newIndex) foundDupe = true;
				}
				if(!foundDupe)
					duplicateIndex = false;
			}

			// it doesn't already point to us... add a handle value to refer to this server
			HandleValue hdlValues[] = {
					new HandleValue(newIndex, Util.encodeString(DOConstants.OBJECT_SERVER_HDL_TYPE),
							repoHandle)
			};
			AddValueRequest addRq = new AddValueRequest(hdlBytes, hdlValues, hdlID);
			try {
				AbstractResponse rs = resolver.getResolver().processRequest(addRq);
				if(rs!=null && rs.responseCode==AbstractMessage.RC_SUCCESS) {
					return;
				} else {
					throw new DOException(DOException.INTERNAL_ERROR,
							"Error updating handle: "+rs);
				}
			} catch (HandleException e) {
				throw new DOException(DOException.INTERNAL_ERROR,
						"Unable to register handle: "+e);
			}
		}
	}



	/**
	 * Load a knowbot from the given input stream.  This knowbot is simply loaded
	 * into the service station and can be used to authenticate users, authorize
	 * users, or perform operations on digital objects.
	 */
	void loadServiceBot(InputStream in)
	throws Exception
	{
		if(in==null) throw new NullPointerException();
		this.knowbotStation.acceptAgent(in);
	}




	//-------------------------------------------------------------------------------
    public void serveRequests() {
	    if(connPool!=null) {
	        // TODO:  Clear existing handler pool
	    }

	    if(DEBUG) System.out.print("Starting DOP connection handlers: ");
	    connPool = new RequestHandlerPool();
	    connPool.setHandlerLife(NUM_CONNECTIONS_PER_HANDLER);
	    for(int threadNum=0; threadNum < NUM_PRELOADED_CONNECTION_HANDLERS; threadNum++) {
	        System.out.print('.');
	        connPool.addHandler(new ConnectionHandler());
	    }
	    if(DEBUG) System.out.println("");
	    System.out.flush();

	    if(DEBUG) System.out.print("Starting BOP operation handlers: ");
	    opHandlerPool = new RequestHandlerPool();
	    opHandlerPool.setHandlerLife(NUM_OPS_PER_HANDLER);
	    for(int i=0; i<NUM_PRELOADED_OP_HANDLERS; i++) 
	        opHandlerPool.addHandler(new OperationRunner());

        if(DEBUG)
	        System.out.println("");
	    System.out.flush();

	    Thread dopListener = null;
	    Thread sslListener = null;

	    try {
	        keepServing = true;

            //System.setProperty("DEBUG", "y");

            // setup the DOP listener socket
            int dopPort = serverConfig.getInt(PORT_KEY, DEFAULT_PORT);
            socket = null;
            if(dopPort > 0) {
              System.err.println("Initializing DOP interface on port "+dopPort);
              if(listenAddress!=null) {
                socket = new ServerSocket(dopPort, DOP_LISTEN_BACKLOG, listenAddress);
              } else {
                socket = new ServerSocket(dopPort, DOP_LISTEN_BACKLOG);
              }

              dopListener = new DOPListener();
              dopListener.start();
            }

            // setup the DOP-over-SSL listener socket
	        int dopSSLPort = serverConfig.getInt(SSL_PORT_KEY, DEFAULT_SSL_PORT);
            sslSocket = null;
	        if(dopSSLPort > 0) {
              System.err.println("Initializing DOP-SSL interface on port "+dopSSLPort);
              if(listenAddress!=null) {
                sslSocket = (SSLServerSocket)sslSocketFactory.createServerSocket(dopSSLPort, DOPSSL_LISTEN_BACKLOG, listenAddress);
              } else {
                sslSocket = (SSLServerSocket)sslSocketFactory.createServerSocket(dopSSLPort, DOPSSL_LISTEN_BACKLOG);
              }
              sslListener = new SSLListener();
              sslListener.start();
	        }

            Server httpServer = startHttpServer();
            
	        if(dopListener!=null) dopListener.join();
	        if(sslListener!=null) sslListener.join();
	        if(httpServer!=null) httpServer.join();  // todo : probably want http Server.join here
	    } catch (Exception e) {
	        e.printStackTrace(System.err);
	    } finally {
	        keepServing = false;
	        try { socket.close(); } catch (Throwable t) {}
	        try { httpSocket.close(); } catch (Throwable t) {}
	    }
	}

    private Server startHttpServer() throws NoSuchAlgorithmException, KeyStoreException, IOException, CertificateException, FileNotFoundException,
            UnrecoverableKeyException, KeyManagementException, Exception {
        // setup an HTTP and/or HTTPS listener
        Server httpServer = null;
        int httpPort = serverConfig.getInt(HTTP_PORT_KEY, DEFAULT_HTTP_PORT);
        int httpsPort = serverConfig.getInt(HTTPS_PORT_KEY, DEFAULT_HTTPS_PORT);
        if(httpPort > 0 || httpsPort > 0) {
          System.setProperty("org.eclipse.jetty.util.log.class", "org.eclipse.jetty.util.log.StdErrLog");
          httpServer = new Server();
          SSLContext sslContext = null;
          
          if(httpsPort>0) {
            String keystorePassStr = serverConfig.getStr("https_keystore_password", null);
            char keystorePass[] = keystorePassStr==null ? null : keystorePassStr.toCharArray();
            String keyPassStr = serverConfig.getStr("https_key_password", null);
            char keyPass[] = keyPassStr==null ? new char[0] : keyPassStr.toCharArray();
            KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
            KeyStore httpsKeyStore = KeyStore.getInstance("JKS");
            httpsKeyStore.load(new FileInputStream(new File(baseDir, HTTPS_KEY_STORE_FILE)), keystorePass);
            kmf.init(httpsKeyStore, keyPass);
            
            sslContext = SSLContext.getInstance("TLS");
            sslContext.init(kmf.getKeyManagers(),
                            new TrustManager[] { new DOSSLTrustManager() },
                            ConnectionEncryption.getRandom());
            SslContextFactory sslContextFactory = new SslContextFactory();
            //sslContextFactory.setNeedClientAuth(true);
            sslContextFactory.setWantClientAuth(true);
            sslContextFactory.setSslContext(sslContext);
            SslSelectChannelConnector sslConnector = new SslSelectChannelConnector(sslContextFactory);
            String httpsHost = serverConfig.getStr(HTTPS_HOSTNAME_KEY, null);
            if(httpsHost!=null) sslConnector.setHost(httpsHost);
            sslConnector.setPort(httpsPort);
            if(listenAddress!=null) sslConnector.setHost(listenAddress.getHostAddress());
            httpServer.addConnector(sslConnector);
          }

          if(httpPort > 0) {
            SelectChannelConnector httpConnector = new SelectChannelConnector();
            httpConnector.setPort(httpPort);
            httpConnector.setMaxIdleTime(30000);
            httpConnector.setRequestHeaderSize(8192);
            httpServer.addConnector(httpConnector);
            if(listenAddress!=null) httpConnector.setHost(listenAddress.getHostAddress());
          }
          
          ContextHandlerCollection contexts = configureServletContainer(httpServer);
          ServletContextHandler context = new ServletContextHandler(ServletContextHandler.SESSIONS);
          context.setContextPath("/");
          HTTPConnectionHandler httpHandler = new HTTPConnectionHandler(thisServer);
          httpHandler.setupServletHolder(context);
          context.addFilter(org.eclipse.jetty.servlets.GzipFilter.class,"/*",null);
          contexts.addHandler(context);
          httpServer.setHandler(contexts);
          
          httpServer.start();
        }
        return httpServer;
    }

    private ContextHandlerCollection configureServletContainer(Server httpServer) {
        ContextHandlerCollection contexts = new ContextHandlerCollection();
          
          File webAppsDir = new File(baseDir,"webapps");
          if((webAppsDir.exists() && webAppsDir.isDirectory()) || webAppsDir.mkdirs()) {
              WebAppProvider webAppProvider = new WebAppProvider() {
                  @Override
                  public ContextHandler createContextHandler(App app) throws Exception {
                      WebAppContext res = (WebAppContext)super.createContextHandler(app);
                      res.addSystemClass("net.cnri.apps.doserver.");
                      res.addSystemClass("net.cnri.dobj.");
                      return res;
                  }
              };
              webAppProvider.setMonitoredDirName(webAppsDir.getAbsolutePath());
              webAppProvider.setParentLoaderPriority(false); 
              webAppProvider.setExtractWars(true);
              File webAppsTempDir = new File(baseDir,"webapps-temp");
              webAppsTempDir.mkdirs();
              webAppProvider.setTempDir(webAppsTempDir);
              webAppProvider.setScanInterval(30);
              // defaults descriptor
              DeploymentManager deploymentManager = new DeploymentManager();
              deploymentManager.setContexts(contexts);
              deploymentManager.setContextAttribute("net.cnri.apps.doserver.Main",this);
              deploymentManager.addAppProvider(webAppProvider);
              httpServer.addBean(deploymentManager);
          }
        return contexts;
    }


	private class DOPListener 
	extends Thread
	{

	    public void run() {
	        while(keepServing && socket!=null) {
	            try {
	                Socket clientSock = socket.accept();
	                ((ConnectionHandler)connPool.getHandler()).handleConnectionAsync(clientSock);
	            } catch (Exception e) {
	                System.err.println("Error accepting DOP connection: "+e);
	            }
	        }
	    }

	}


	private class SSLListener 
	extends Thread
	{

		public void run() {
			while(keepServing && sslSocket!=null) {
				try {
					Socket clientSock = sslSocket.accept();
					((ConnectionHandler)connPool.getHandler()).handleConnectionAsync(clientSock);
				} catch (Exception e) {
					System.err.println("Error accepting DOP/SSL connection: "+e);
          e.printStackTrace(System.err);
				}
			}
		}
	}


  private class HTTPSListener
  extends Thread
  {

      public void run() {
          while(keepServing && sslSocket!=null) {
              try {
                  Socket clientSock = sslSocket.accept();
                  ((ConnectionHandler)connPool.getHandler()).handleConnectionAsync(clientSock);
              } catch (Exception e) {
                  System.err.println("Error accepting DOP/SSL connection: "+e);
        e.printStackTrace(System.err);
              }
          }
      }
  }

	public boolean isListening() {
		try {
			if(socket!=null)
				return socket.isBound();
		} catch (Exception e) {
			System.err.println("error checking socket for listen status: "+e);
		}
		return false;
	}

	public void shutdown() 
	throws Exception
	{
		stopServing();

		if(replicator!=null) replicator.shutdown();
		replicator = null;

		java.util.Enumeration en = knowbotStation.getCurrentAgents();
		while(en.hasMoreElements()) {
			try {
				Object agent = en.nextElement();
				((KnowbotWrapper)agent).stop();
			} catch(Throwable t) {
				System.err.println("Error stopping agent: "+t);
			}
		}

		storage.getTransactionQueue().shutdown();

		if(logger!=null) logger.shutdown();
		errLogger.shutdown();
	}

	public void stopServing() 
	throws IOException
	{
		keepServing = false;
		if(socket!=null) {
			socket.close();
			socket = null;
		}
		if(sslSocket!=null) {
			sslSocket.close();
			sslSocket = null;
		}
		if(httpSocket!=null) {
			httpSocket.close();
			httpSocket = null;
		}
	}


	/** This method can be called by an outside object when this repository is run
	 * in embedded mode.  No built-in authentication is performed and the 
	 * operation is invoked with the caller set to the server itself. */
	public void performOperation(String objectID, String operationID, HeaderSet params,
			InputStream in, OutputStream out)
	throws Exception
	{
		if(params==null) params = new HeaderSet();
		performOperation(new RequestContext(null, getServerID(),
                                        operationID,
                                        objectID, params), in, out);
	}


  public void performOperation(DOServerOperationContext context, InputStream in, OutputStream out)
	throws Exception
	{
		StorageProxy storage = context.getStorage();
		if(!storage.doesObjectExist()) {
			Authorizer.AuthorizationInfo auth = authorizer.forwardingIsAllowed(context, null);
			boolean canForward = auth.canForwardOperation();
			//logger.log(context.getLogEntry(dateFmt, canForward));
			if(!canForward) {
				HeaderSet response = new HeaderSet("response");
				response.addHeader("status", "error");
				response.addHeader("message", "Object "+context.getTargetObjectID()+" not located here");
				response.addHeader("code", DOException.NO_SUCH_OBJECT_ERROR);
				response.writeHeaders(out);
				out.close();
				return ;
			}

			performRemoteOperation(context, auth, in, out);
		} else {
			performLocalOperation(context, false, in, out);
		}
	}


	/** Forward the operation to the repository that is responsible for the object,
	 * if the client is permitted to do so. */
	public void performRemoteOperation(DOServerOperationContext context, Authorizer.AuthorizationInfo auth,
                                     InputStream in, OutputStream out)
	throws Exception
	{
		DOClient fwdClient = auth.useRepoID() ? 
      doClient : context.getDOClientWithForwardedAuthentication(context.getCallerID());

		StreamPair io = null;
		try {
			// attempt to perform the operation using the 
			io = fwdClient.performOperation(auth.getEndpointRepositoryID(),
					context.getTargetObjectID(), 
					context.getOperationID(),
					context.getOperationHeaders());
		} catch (DOException e) {
			HeaderSet response = new HeaderSet("response");
			response.addHeader("status", "error");
			response.addHeader("message", e.getMessage());
			response.addHeader("code", e.getErrorCode());
			response.writeHeaders(out);
			out.close();
			return;
		}

		HeaderSet response = new HeaderSet("response");
		response.addHeader("status", "success");
		response.writeHeaders(out);
		out.flush();

		// spawn another thread to deal with the outbound bytes
		// TODO: Replace this with an non-blocking write, perhaps using Channels
		Thread outboundForwarder = new OutboundForwarder(in, io.getOutputStream());
		outboundForwarder.start();

		// forward the incoming bytes in this thread
		IOForwarder.forwardStream(io.getInputStream(), out);
		out.close();

		// wait for the outbound bytes to finish
		outboundForwarder.join();
		return;
	}

	private class OutboundForwarder
	extends Thread
	{
		private InputStream in;
		private OutputStream out;

		OutboundForwarder(InputStream in, OutputStream out) {
			this.in = in;
			this.out = out;
		}

		public void run() {
			try {
				IOForwarder.forwardStream(in, out);
				out.close();
			} catch (Throwable t) {
				System.err.println("error reading forwarded stream: "+t);
				try { out.close(); } catch (Throwable t2) {}
			} 
		}
	}


	/** Invoke the given operation in the repository after checking that the client has
	 * permission to invoke the given operation */
	public void performLocalOperation(DOServerOperationContext context, boolean skipAuthorization, 
                                     InputStream in, OutputStream out)
  	throws Exception
	{
		boolean authorized = skipAuthorization || authorizer.operationIsAllowed(context);
		if(logger!=null) logger.log(context.getLogEntry(dateFmt, authorized));
		if(!authorized) {
			in.close();
			HeaderSet response = new HeaderSet("response");
			response.addHeader("status", "error");
			response.addHeader("message", "operation not allowed");
			response.addHeader("code", DOException.PERMISSION_DENIED_ERROR);
			response.writeHeaders(out);
			out.close();
			return;
		}
		
		String opID = context.getOperationID();
		Object operators[] =
			knowbotStation.getCollaborationHub().lookupEntity(null,
					null,
					DOOperation.class);
		if(opID.equalsIgnoreCase(DOConstants.LIST_OPERATIONS_OP_ID)) {
		  in.close();
		  HeaderSet response = new HeaderSet("response");
		  response.addHeader("status", "success");
		  response.writeHeaders(out);
		  
		  OutputStreamWriter w = new OutputStreamWriter(out, "UTF8");
		  Hashtable alreadySentOps = new Hashtable();
		  try {
		    w.write(DOConstants.LIST_OPERATIONS_OP_ID);
		    w.write("\n");
				for(int i=0; operators!=null && i<operators.length; i++) {
					if(operators[i]==null) continue;
					String ops[] = ((DOOperation)operators[i]).listOperations(context);
					if(ops==null) continue;
					for(int j=0; j<ops.length; j++) {
						if(ops[j]==null) continue;
						if(!alreadySentOps.containsKey(ops[j].toLowerCase())) {
							w.write(ops[j]);
							w.write("\n");
							alreadySentOps.put(ops[j].toLowerCase(), "");
						}
					}
				}
			} finally {
				w.flush();
				out.close();
			}
		} else {
			// find a knowbot to handle this operation
			// a standard operation is being invoked... find an operator to handle it
			DOOperation op = null;
			for(int i=0; operators!=null && i<operators.length; i++) {
				DOOperation tmpOp = (DOOperation)operators[i];
				if(tmpOp.canHandleOperation(context)) {
					op = tmpOp;
					break;
				}
			}
			
			if(op==null) {
				HeaderSet response = new HeaderSet("response");
				response.addHeader("status", "error");
				response.addHeader("message", "Operation '"+opID+"' Not Available");
				response.addHeader("code", DOException.OPERATION_NOT_AVAILABLE);
				response.writeHeaders(out);
				out.close();
				return;
			}
			
			op.performOperation(context, in, out);
		}
	}


	/** ConnectionHandler objects run in their own thread and are responsible for
	 * managing a single connection to the server.  Each connection may have
	 * numerous operation invocations which is why each connection handler has
	 * its own RequestHandlerPool to keep track of threads for operation invocation
	 * requests. */
	private class ConnectionHandler
	implements Runnable, RequestHandler, DOConnectionListener
	{
		private boolean keepRunning = true; // should the handler thread keep going
		private boolean isRunning = false;  // is the handler thread busy with a request
		private Thread thread = null;
		private Object waitLock = new Object();
		private Socket socket = null;
		private DOServerConnection doServer = null;
		private HashMap fwdClients = new HashMap();
		private DOClient forwardingClient = null;

		private int invocationCount = 0;

		ConnectionHandler() {
			thread = new Thread(this);
			thread.start();
		}

		/** Resets the state of the handler and prepares it to handle
		 * another request. */
		public synchronized void resetState() {
			try { if(doServer!=null) doServer.close(); } catch (Throwable t) {}
			try {
				for(Iterator iter=fwdClients.keySet().iterator(); iter.hasNext(); ) {
					String key = (String)iter.next();
					((DOClient)fwdClients.get(key)).closeAllConnections();
				}
			} catch (Throwable t) {
				System.err.println("Error closing forwarded connections: "+t);
				t.printStackTrace(System.err);
			}
			fwdClients.clear();
			doServer = null;
			thread.setName("Unused Connection Handler");
		}

		/** Construct another handler that can be used to manage connections.
		 * This should do something like clone() except this should fork a
		 * new thread. */
		public RequestHandler newHandler() {
			return new ConnectionHandler();
		}

		/** Return the IP address of the other side of the connection */
		public InetAddress getClientIP() {
			return socket.getInetAddress();
		}

		/** Shut down this handler.  This should stop the handler thread and
		 * clean up any open resources. */
		public void deactivate() {
			keepRunning = false;
			try { if(doServer!=null) doServer.close(); } catch (Throwable t) {}
			doServer = null;
			synchronized(waitLock) { waitLock.notify(); }
		}

		/** Get a count of how many times this handler has been invoked. */
		public int getInvocationCount() {
			return invocationCount;
		}

		/** Pass the socket off to a handler thread which will manage the 
		 * connection. */
		void handleConnectionAsync(Socket sock) {
			synchronized(waitLock) {
				this.socket = sock;
				isRunning = true;
				waitLock.notify();
			}
		}

		public void run() {
			while(keepRunning) {
				synchronized(waitLock) {
					while(!isRunning && keepRunning) {
						try {
							waitLock.wait();
						} catch (Exception e) {
							System.err.println("Error waiting for request in connection handler: "+e);
						}
					}
				}

				if(!keepRunning) continue;

				try {
					this.doServer = new DOServerConnection(myID, socket, this, Main.this.delegationClient);
					thread.setName("Connection Handler: "+socket.getInetAddress().getHostAddress());

					while(doServer.isOpen()) {
						// TODO: set a time-out for connections that have not had 
						// activity in a certain time period to be closed
						Thread.sleep(10000);
					}
				} catch (Exception e) {
					System.err.println("protocol/connection error: "+e+"; closing connection");
					//e.printStackTrace(System.err);
				} finally {
					try {
						if(socket!=null) socket.close();
					} catch (Throwable e) {
						System.err.println("Error closing server socket: "+e);
						e.printStackTrace(System.err);
					}
					isRunning = false;
					connPool.returnHandler(this);
				}
			}
		}

		/** Acquires an OperationRunner from the pool to process an operation
		 * over the given newly created channel. */
		public void channelCreated(StreamPair pair) {
			// get a handler to take care of the operation on the new channel
			invocationCount++;
			OperationRunner runner = (OperationRunner)opHandlerPool.getHandler();
			runner.processOperationAsynchronously(doServer, pair);
		}
	}

	/** This class is responsible for converting the incoming request message
	 * in an operation invocation.  */
	private class OperationRunner
	implements Runnable, RequestHandler
	{
		private Thread thread;
		private int invocationCount = 0;
		private boolean keepRunning = true;
		private boolean isRunning = false;

		private HeaderSet headers = null;
		private DOServerConnection doServer = null;
		private StreamPair pair = null;
		private long startTime = 0;
		private Object waitLock = new Object();

		OperationRunner() {
			thread = new Thread(this);
			thread.start();
		}

		/** Resets the state of the handler and prepares it to handle another request */
		public void resetState() {
			headers = null;
			doServer = null;
			try { if(pair!=null) pair.close(); } catch (Throwable t) {}
			pair = null;
			thread.setName("Unused Operation Handler");
		}

		/** Construct another handler that can be used to process requests.
        This should do something like clone() except this should fork a
        new thread. */
		public RequestHandler newHandler() {
			return new OperationRunner();
		}

		/** Shut down this handler.  This should stop the handler thread and
      clean up any open resources. */
		public void deactivate() {
			keepRunning = false;
		}

		/** Get a count of how many times this handler has been invoked. */
		public int getInvocationCount() {
			return invocationCount;
		}


		/** Hand off the given operation to a separate thread */
		void processOperationAsynchronously(DOServerConnection doServer,
				StreamPair io)
		{
			synchronized (waitLock) {
				this.doServer = doServer;
				this.pair = io;
				this.headers = null;
				isRunning = true;
				waitLock.notify();
			}
		}

		void processOperationSynchronously(DOServerConnection doServer,
				StreamPair io,
				HeaderSet requestInfo)
		{
			synchronized (waitLock) {
				this.doServer = doServer;
				this.pair = io;
				this.headers = requestInfo;
				isRunning = true;
				processOperation();
			}
		}

		private void processOperation() {
			// perform the operation!
            InputStream in = null;
            OutputStream out = null;
            RequestContext reqContext = null;
			try {
	            invocationCount++;
	            thread.setName("Operation Handler: initializing");

	            in = pair.getInputStream();
				out = pair.getOutputStream();
				HeaderSet headers = this.headers;
				if(headers==null) { // get the headers if they weren't already provided
					headers = new HeaderSet();
					headers.readHeaders(in);
				}

				thread.setName("Operation Handler: "+headers);
				String operationID = headers.getStringHeader("operationid", "");
				String callerID = headers.getStringHeader("callerid", "");
				String objectID = headers.getStringHeader("objectid", "");
				HeaderSet parameters = headers.getHeaderSubset("params");
				if(auditLogEnabled) {
					StreamPair auditPair = auditLog.recordOperation(headers, in, out);
					in = auditPair.getInputStream();
					out = auditPair.getOutputStream();
				}
				reqContext = new RequestContext(doServer, callerID, operationID,
						objectID, parameters);
				thread.setName("Operation Handler: "+reqContext);
				pair.setName(String.valueOf(reqContext));

				performOperation(reqContext, in, out);
			} catch (Exception e) {
				e.printStackTrace(System.err);
			} finally {
				isRunning = false;
				try { out.close(); } catch (Throwable e) {}
				try { in.close(); } catch (Throwable e) {}
				opHandlerPool.returnHandler(this);
			}
		}

		public void run() {
			while(keepRunning) {
				synchronized(waitLock) {
					while(!isRunning && keepRunning) {
						try {
							waitLock.wait();
						} catch (Exception e) {
							System.err.println("operation runner: error waiting for request");
						}
					}
				}
				if(!keepRunning) continue;

				processOperation();
			}
		}
	}

	private class RequestContext
	extends DOServerOperationContext
	{
	    private String callerID;
	    private String opID;
	    private String objectID;
	    private HeaderSet params;
	    private ConcreteStorageProxy storageProxy;
	    private DOServerConnection doServer;
	    private StringBuffer logEntryStr = new StringBuffer();
	    private RequestContext(DOServerConnection doServer, String callerID,
	            String opID, String objectID, HeaderSet params) {
	        this.callerID = callerID;
	        this.opID = opID;
	        this.objectID = objectID;
	        this.params = params;
	        this.doServer = doServer;
	        this.storageProxy = null;
	    }

	    public InetAddress getClientAddress() {
	        if(doServer==null) return null;
	        return doServer.getSocket().getInetAddress();
	    }

	    /**
	     * Requests verification of the callers identity.  Returns true iff the
	     * caller's identity has been verified.
	     */
	    public synchronized boolean authenticateCaller() {
	        if(doServer==null) return true;
	        boolean authenticated = false;
	        try {
	            authenticated = doServer.authenticateClient(callerID,getStoredPasswordForUser(callerID));
	        } catch (DOException e) {
	            System.err.println("Error authenticating: "+e);
	            e.printStackTrace(System.err);
	        }
	        return authenticated;
	    }
	    /**
	     * Returns a list of unverified IDs that the client claims as credentials.
	     * Note: These IDs are not yet verified and the caller should call
	     * authenticateCredential() with any credential IDs for whom they
	     * assign any meaning.
	     * Note2: This should only be called after authenticateClient returns.
	     */
	    public String[] getCredentialIDs() {
	        if(doServer==null) return null;
	        return doServer.getCredentialIDs();
	    }
	    /**
	     * Verify that this client has been granted a credential by the identified entity.
	     * The given credentialID is expected to have come from the list returned by
	     * getCredentialIDs().  Returns true iff a verified credential from credentialID
	     * was granted to this client.
	     */
	    public boolean authenticateCredential(String credentialID) {
	        if(doServer==null) return true;
	        return doServer.authenticateCredential(credentialID);
	    }
	    /**
	     * Performs the specified operation with the identity of the caller.
	     * If the serverID is null then the DO client that performs this request
	     * will resolve the objectID to find the server.
	     */
	    public void performOperation(String serverID, String objectID,
	            String operationID, HeaderSet params,
	            InputStream input, OutputStream output)
	                    throws DOException
	                    {
	        RequestContext derivedCtx = new RequestContext(doServer, callerID, operationID,
	                objectID, params);
	        try {
	            thisServer.performOperation(derivedCtx, input, output);
	        } catch (Exception e) {
	            e.printStackTrace(System.err);
	            if(e instanceof DOException) throw (DOException)e;
	            throw new DOException(DOException.SERVER_ERROR,
	                    "Error invoking derived operation: "+e);
	        }
	                    }
	    /**
	     * Returns the identity of the caller.
	     */
	    public String getCallerID() { return callerID; }
	    /**
	     * Returns the operation that the caller attempted to invoke.
	     */
	    public String getOperationID() { return opID; }
	    /**
	     * Returns the object on which the caller is invoke the operation.
	     */
	    public String getTargetObjectID() { return objectID; }
	    /**
	     * Returns the identity of this server.
	     */
	    public String getServerID() { return serverID; }
	    /**
	     * Returns the set of headers that were included with the operation request
	     */
	    public HeaderSet getOperationHeaders() { return params; }
	    /**
	     * Returns on object that allows operators to access the storage
	     * system for the current object.
	     */
	    public StorageProxy getStorage() {
	        // only construct the storage proxy if it is requested
	        synchronized(this) {
	            if(storageProxy==null) {
	                HeaderSet txnMetadata = new HeaderSet();
	                txnMetadata.addHeader("callerid", callerID);
	                txnMetadata.addHeader("operationid", opID);
	                txnMetadata.addHeader("params", params);
	                storageProxy =
	                        new ConcreteStorageProxy(storage, myID.getID(), objectID, txnMetadata);
	            }
	        }
	        return storageProxy;
	    }

	    /**
	     * Inserts an object into the connection-level information table.  This
	     * causes subsequent calls to getConnectionMapping() on the same connection
	     * (but not necessarily the same operation) with the same key to return
	     * the given data value.
	     */
	    public void setConnectionMapping(Object mappingKey, Object mappingData) {
	        if(doServer==null) return;
	        this.doServer.setConnectionMapping(mappingKey, mappingData);
	    }

	    /**
	     * Returns the object from the connection-level information table that had
	     * previously been used as the mappingData parameter for a call to
	     * setConnectionMapping with the given mappingKey.  If no such object
	     * had been inserted into the table then this will return null.
	     */
	    public Object getConnectionMapping(Object mappingKey) {
	        if(doServer==null) return null;
	        return this.doServer.getConnectionMapping(mappingKey);
	    }

	    /** Return a DOClient instance that forwards any authentication challenges back to
	     * the DOP client on the other end of this connection.
	     */
	    public synchronized DOClient getDOClientWithForwardedAuthentication(String clientID) {
	        String mappingKey = "doclient-"+clientID;
	        DOClient client = (DOClient)doServer.getConnectionMapping(mappingKey);
	        if(client!=null) return client;

	        client = new DOClient(new ProxiedAuthentication(clientID, doServer));
	        doServer.setConnectionMapping(mappingKey, client);
	        return client;
	    }

	}
}