package net.cnri.apps.doserver;

import net.cnri.apps.doserver.HTTPConnectionHandler.HTTPRepositoryInterface;
import net.cnri.do_api.DataElement;
import net.cnri.do_api.DigitalObject;
import net.cnri.do_api.Repository;
import net.cnri.dobj.DOException;
import net.cnri.dobj.HeaderSet;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.util.*;

public class ObjectFactoryServlet extends HttpServlet {
  private Main serverMain;
  private HTTPConnectionHandler httpConnectionHandler;
  private final static String FACTORY_DIR_NAME = "factory";
  private Map<String, Properties> props;
  private final static String REQUIRE = "require";
  private final static String MAP = "map";
  private final static String DEFINE = "define";

  public ObjectFactoryServlet(Main main, HTTPConnectionHandler httpConnectionHandler) {
    super();
    this.serverMain = main;
    this.httpConnectionHandler = httpConnectionHandler;
  }

  @Override
  protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
    //URL requestURI = new URL(request.getRequestURI());
    //System.err.println("URL:"+request.getRequestURL());
    //System.err.println("URI:"+request.getRequestURI());
    
    //URL requestUrl = new URL(new String(request.getRequestURL()));
    //System.err.println("PATH:"+requestUrl.getPath());
    
    //validate the path; length 2, properties file for <last_path_seg>.properties
    String path = request.getRequestURI();
    if(path.endsWith("/")) {
      path = path.substring(0, path.length() - 1);
    }
    if(path.startsWith("/")) {
      path = path.substring(1);
    }
    String segments[] = path.split("/");
    
    if(segments.length != 2) {
      response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
      returnError("Unsupported path:" + request.getRequestURI(), null, response);
      return;
    } 

    //Kick off the processing train
    loadProps(segments[1], request, response);
  }
  
  private void returnError(String message, Exception e, HttpServletResponse response) {
    try {
      Template t = httpConnectionHandler.getTemplate("error_msg.html");
      HashMap errorInfo = new HashMap();
      if(message != null) {
        errorInfo.put("message", message);
      }
      if(e != null) {
        errorInfo.put("error", message);
      }
      t.merge(new VelocityContext(errorInfo),response.getWriter());
    } catch (Exception e2) {
      e2.printStackTrace(System.err);
    }
  } 
  
  private void loadProps(String prefix, HttpServletRequest request, HttpServletResponse response) {
    props = new HashMap<String, Properties>();
    //These would be the mappings of http param names to their repo attributes
    File map = new File(serverMain.getBaseFolder(), FACTORY_DIR_NAME+"/"+prefix+"."+MAP+".properties");
    //System.err.println(map);
    Properties fileProps;
    if(!map.exists()) {
      response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      returnError("File does not exist:" + map, null, response);
      return;
    } else {
      fileProps = new Properties();
      try {
        fileProps.load(new InputStreamReader(new FileInputStream(map), "UTF-8"));
        props.put(MAP, fileProps);
      } catch (Exception e) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        returnError(null, e, response);
        return;
      }
    }

    //Check for require, basically just ends up be a list
    File require = new File(serverMain.getBaseFolder(), FACTORY_DIR_NAME+"/"+prefix+"."+REQUIRE+".properties");
    //System.err.println(require);
    if(!require.exists()) {
      props.put(REQUIRE, new Properties());
    } else {
      fileProps = new Properties();
      try {
        fileProps.load(new InputStreamReader(new FileInputStream(require), "UTF-8"));
        props.put(REQUIRE, fileProps);
      } catch (Exception e) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        returnError(null, e, response);
        return;
      }
    }
    
    //Check and load predefined values
    File define = new File(serverMain.getBaseFolder(), FACTORY_DIR_NAME+"/"+prefix+"."+DEFINE+".properties");
    if(!define.exists()) {
      props.put(DEFINE, new Properties());
    } else {
      fileProps = new Properties();
      try {
        fileProps.load(new InputStreamReader(new FileInputStream(define), "UTF-8"));
        props.put(DEFINE,  fileProps);
      } catch (Exception e) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        returnError(null, e, response);
        return;
      }
    }

    //check that any requires are defined in the map
    for(Map.Entry param : props.get(REQUIRE).entrySet()) {
      if(!props.get(MAP).containsKey(param.getKey())) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        returnError("Unresolved param:"+param.getKey(), null, response);
        return;
      } else if(request.getParameter(param.getKey().toString()) == null) {
        response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
        returnError("Missing required param:"+param.getKey(), null, response);
        return;
      }
    }
    
    //check that any defines are defined in the map
    for(Map.Entry param : props.get(DEFINE).entrySet()) {
      if(!props.get(MAP).containsKey(param.getKey())) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        returnError("Unresolved defined param:"+param.getKey(), null, response);
        return;
      }
    }
    validateRequest(request, response);
  }
  
  private void validateRequest(HttpServletRequest request, HttpServletResponse response) {
    //go over the parameters in the request make sure all are permitted
    Enumeration<String> e = request.getParameterNames();
    String param;

    while(e.hasMoreElements()) {
      param = e.nextElement();
      if(!props.get(MAP).containsKey(param)) {
        response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
        returnError("Param not allowed:"+param, null, response);
        return;
      }
    }
    
    performRequest(request, response);
  }
  
  private void performRequest(HttpServletRequest request, HttpServletResponse response) {
    try {
      ObjectComponent builder = new ObjectComponent();
      String param, resolvedParam, paramValue;
      DoComponent component;
      for(Map.Entry entry : props.get(DEFINE).entrySet()) {
        //Add all predefined values
        param = entry.getKey().toString();
        paramValue = entry.getValue().toString();
        resolvedParam = props.get(MAP).get(param).toString();
        component = resolveDoComponent(resolvedParam, builder);
        component.putValue(paramValue);
      }
      
      Enumeration<String> params = request.getParameterNames();
      while(params.hasMoreElements()) {
        //Add all of the request defined values
        param = params.nextElement();
        paramValue = request.getParameter(param);
        resolvedParam = props.get(MAP).get(param).toString();
        component = resolveDoComponent(resolvedParam, builder);
        component.putValue(paramValue);
      }
      
      //Build the digital object
      //System.err.println("Object");
      HeaderSet atts = new HeaderSet();
      for(AttComponent attComponent : builder.atts) {
        //System.err.println("Object/att/"+attComponent.key+":"+attComponent.value);
        atts.addHeader(attComponent.key, attComponent.value);
      }
      
      Repository repo = new HTTPRepositoryInterface(serverMain, request, serverMain.getServerID());
      DigitalObject factoryObject = repo.createDigitalObject(null);
      factoryObject.setAttributes(atts);
      
      DataElement dataElement = null;
      for(Map.Entry<String, ElComponent> entry : builder.els.entrySet()) {
        //Add all of the elements that might be defined
        dataElement = factoryObject.getDataElement(entry.getKey());
        //System.err.println("Object/el/"+entry.getKey()+"/data:"+entry.getValue().data);
        dataElement.write(new ByteArrayInputStream(entry.getValue().data.getBytes("UTF-8")));

        //Add all of the data element attributes
        atts = new HeaderSet();
        for(AttComponent attComponent : entry.getValue().atts) {
          //System.err.println("Object/el/"+entry.getKey()+"/att/"+attComponent.key+":"+attComponent.value);
          atts.addHeader(attComponent.key, attComponent.value);
        }
        dataElement.setAttributes(atts);
      }

      response.addHeader("ObjectID", factoryObject.getObjectID());

      finishCreated(response);
    } catch (DOException e) {
      response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      returnError(null, e, response);
      return;
    } catch (IOException e) {
      response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      returnError(null, e, response);
      return;
    }
  }

  private DoComponent resolveDoComponent(String path, DoComponent component) {
    //System.err.println("path:"+path);
    if("".equals(path)) {
      //end of path so return;
      return component;
    }
    
    DoComponent comp = null;
    if(path.startsWith("/att/") || path.startsWith("att/")) {
      if(path.startsWith("/")) {
        path = path.substring(1);
      }
      path = path.substring(4);
      //System.err.println("ATT:"+path);
      comp = component.getAtt(path);
      path = "";
    }

    if(path.startsWith("/el/") || path.startsWith("el/")) {
      if(path.startsWith("/")) {
        path = path.substring(1);
      }
      path = path.substring(3);
      //System.err.println("EL:"+path);
      comp = component.getElement(path);
      //System.err.println(comp);
      if(path.indexOf("/") != -1) {
        path = path.substring(path.indexOf("/"));
      } else {
        //just the element name so empty the string
        path = "";
      }
    }

    return resolveDoComponent(path, comp);
  }
  
  private void finishCreated(HttpServletResponse response) {
    response.setStatus(HttpServletResponse.SC_CREATED);
    return;
  }
  
  private static interface DoComponent {
    public void putValue(String value);
    public DoComponent getElement(String elName);
    public DoComponent getAtt(String attName);
  }

  /** Handle the attributes **/
  private static class AttComponent implements DoComponent {
    String key;
    String value;
    AttComponent(String key) {
      this.key = key;  
    }
    
    public DoComponent getElement(String elName) {
      throw new RuntimeException("Not valid element:"+elName);
    }
    
    public DoComponent getAtt(String attName) {
      throw new RuntimeException("Not valid attribute:"+attName);
    }
    
    public void putValue(String value) {
      this.value = value;  
    }
    
  }

  /** Handle the elements **/
  private static class ElComponent implements DoComponent {
    List<AttComponent> atts;
    String data;
    
    ElComponent() {
      data = "";
      atts = new ArrayList<AttComponent>();
    }
    
    public DoComponent getElement(String elName) {
      throw new RuntimeException("Not a valid element:"+elName);
    }
    
    public DoComponent getAtt(String attName) {
      DoComponent doAtt = null;
      for(AttComponent att : atts) {
        if(att.key.equals(attName)) {
          doAtt = att;
          break;
        }
      }
      if(doAtt == null) {
        doAtt = new AttComponent(attName);
        atts.add((AttComponent)doAtt);
      }
      return doAtt;
    }
    
    public void putValue(String value) {
      data = value;
    }
  }


  /** Handle the base object **/
  private static class ObjectComponent implements DoComponent {
    List<AttComponent> atts;
    Map<String, ElComponent> els;
    
    ObjectComponent() {
      atts = new ArrayList<AttComponent>();
      els = new HashMap<String, ElComponent>();
    }
    
    public DoComponent getElement(String elPath) {
      //assume we are passed <elementname>[/other stuff]
      String elName = null;
      if(elPath.indexOf("/") != -1) {
        elName = elPath.substring(0, elPath.indexOf("/"));
      } else {
        elName = elPath;
      }

      ElComponent el = null;
      if(!els.containsKey(elName)) {
        el = new ElComponent();
        els.put(elName, el);
      } else {
        el = els.get(elName);
      }
      return el;
    }

    public DoComponent getAtt(String attName) {
      AttComponent doAtt = null;
      for(AttComponent att : atts) {
        if(att.key.equals(attName)) {
          doAtt = att;
          break;
        }
      }
      if(doAtt == null) {
        doAtt = new AttComponent(attName);
        atts.add(doAtt);
      }
      return doAtt;
    }
    
    public void putValue(String value) {
      throw new RuntimeException("Improper value for ObjectComponent:"+value);
    }
  }
}
  
