package net.cnri.apps.doserver.operators;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import net.cnri.apps.doserver.Main;
import net.cnri.do_api.DOKeyRing;
import net.cnri.do_api.DataElement;
import net.cnri.do_api.DigitalObject;
import net.cnri.dobj.DOConstants;
import net.cnri.dobj.DOException;
import net.cnri.dobj.HeaderItem;
import net.cnri.dobj.HeaderSet;
import net.cnri.dobj.StorageProxy;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.KeywordAnalyzer;
import org.apache.lucene.analysis.PerFieldAnalyzerWrapper;
import org.apache.lucene.analysis.WhitespaceAnalyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.DateTools;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.MatchAllDocsQuery;
import org.apache.lucene.search.PhraseQuery;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.BooleanClause.Occur;
import org.apache.lucene.util.Version;
import org.apache.nutch.metadata.Metadata;
import org.apache.nutch.parse.Parse;
import org.apache.nutch.parse.ParseData;
import org.apache.nutch.parse.Parser;
import org.apache.nutch.protocol.Content;

public class DefaultIndexBuilder implements IndexBuilder {
    private static final int VERSION = 1;
    
    private static final String[] STOP_WORDS_EN_ARR =  {
        "a"       , "and"     , "are"     , "as"      ,
        "at"      , "be"      , "but"     , "by"      ,
        "for"     , "if"      , "in"      , "into"    ,
        "is"      , "it"      , "no"      , "not"     ,
        "of"      , "on"      , "or"      , "s"       ,
        "such"    , "t"       , "that"    , "the"     ,
        "their"   , "then"    , "there"   , "these"   ,
        "they"    , "this"    , "to"      , "was"     ,
        "will"    ,
        "with"
    };
    private static final Set<String> STOP_WORDS_EN = new HashSet<String>();
    static {
        for(String s : STOP_WORDS_EN_ARR) STOP_WORDS_EN.add(s);
    }

    private static final String[] SEARCH_FIELDS = {
        "eltext_content",
        "objatt_notes",
        "eltitle_content",
        "elatt_content_filename",
    };

    private Main serverMain;
    private org.apache.hadoop.conf.Configuration nutchConfig;
    private HashMap nutchParsers;
    /** The analyzer used by the IndexBuilder. */
    protected PerFieldAnalyzerWrapper analyzer;

    private volatile DOKeyRing keyring;
    
    public boolean DEBUG = false;
    
    public DefaultIndexBuilder() {}

    public void init (Main serverMain) {
        this.serverMain = serverMain;

        PerFieldAnalyzerWrapper analyzerWrapper = new PerFieldAnalyzerWrapper(new StandardAnalyzer(Version.LUCENE_30,STOP_WORDS_EN));
        analyzerWrapper.addAnalyzer("id",new KeywordAnalyzer());
        analyzerWrapper.addAnalyzer("repoid",new KeywordAnalyzer());
        analyzerWrapper.addAnalyzer("indexVersion",new WhitespaceAnalyzer());
        analyzer = analyzerWrapper;

        // initialize the nutch parser plugins
        nutchConfig = new org.apache.hadoop.conf.Configuration();
        nutchConfig.setInt("file.content.limit", 65536);
        nutchConfig.set("mime.types.file", "net/cnri/apps/doserver/operators/lucene-mime-types.xml");
        nutchConfig.set("plugin.folders", serverMain.getBaseFolder().getPath() + File.separator + "index" + File.separator + "plugins");

        nutchParsers = new HashMap();

        Parser msword = new org.apache.nutch.parse.msword.MSWordParser();
        Parser text = new org.apache.nutch.parse.text.TextParser();
        Parser pdf = new org.apache.nutch.parse.pdf.PdfParser();
        Parser rss = new org.apache.nutch.parse.rss.RSSParser();
        Parser msexcel = new org.apache.nutch.parse.msexcel.MSExcelParser();
        Parser mspp = new org.apache.nutch.parse.mspowerpoint.MSPowerPointParser();
        Parser ootext = new org.apache.nutch.parse.oo.OOParser();
        Parser html = new org.apache.nutch.parse.html.HtmlParser();
        Parser zip = new org.apache.nutch.parse.zip.ZipParser();
        //Parser shockwave = new org.apache.nutch.parse.swf.SWFParser();

        msword.setConf(nutchConfig);
        text.setConf(nutchConfig);
        pdf.setConf(nutchConfig);
        rss.setConf(nutchConfig);
        msexcel.setConf(nutchConfig);
        mspp.setConf(nutchConfig);
        ootext.setConf(nutchConfig);
        try { html.setConf(nutchConfig); } catch (Exception e) {/* grrr */ }
        zip.setConf(nutchConfig);

        nutchParsers.put("application/java", text);
        nutchParsers.put("application/pdf", pdf);
        nutchParsers.put("application/postscript", pdf);
        nutchParsers.put("application/msword", msword);
        nutchParsers.put("application/rss+xml", rss);
        nutchParsers.put("application/vnd.ms-excel", msexcel);
        nutchParsers.put("application/vnd.ms-powerpoint", mspp);
        nutchParsers.put("application/vnd.oasis.opendocument.text", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.text-template", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.text-master", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.text-web", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.presentation", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.presentation-template", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.spreadsheet", ootext);
        nutchParsers.put("application/vnd.oasis.opendocument.spreadsheet-template", ootext);
        nutchParsers.put("application/vnd.sun.xml.calc", ootext);
        nutchParsers.put("application/vnd.sun.xml.calc.template", ootext);
        nutchParsers.put("application/vnd.sun.xml.impress", ootext);
        nutchParsers.put("application/vnd.sun.xml.impress.template", ootext);
        nutchParsers.put("application/vnd.sun.xml.writer", ootext);
        nutchParsers.put("application/vnd.sun.xml.writer.template", ootext);
        nutchParsers.put("application/vnd.wap.wbxml", text);
        nutchParsers.put("application/vnd.wap.wmlc", text);
        nutchParsers.put("application/vnd.wap.wmlscriptc", text);
        nutchParsers.put("application/xhtml+xml", text);
        nutchParsers.put("application/x-javascript", "parse-js");
        nutchParsers.put("application/x-kword", msword);
        nutchParsers.put("application/x-kspread", msexcel);
        nutchParsers.put("application/x-latex", text);
        nutchParsers.put("application/x-netcdf", text);
        nutchParsers.put("application/x-sh", text);
        //nutchParsers.put("application/x-shockwave-flash", shockwave);
        nutchParsers.put("application/x-tcl", text);
        nutchParsers.put("application/x-tex", text);
        nutchParsers.put("application/x-texinfo", text);
        nutchParsers.put("application/x-troff", text);
        nutchParsers.put("application/x-troff-man", text);
        nutchParsers.put("application/x-troff-me", text);
        nutchParsers.put("application/x-troff-ms", text);
        nutchParsers.put("application/zip", zip);
        nutchParsers.put("message/news", text);
        nutchParsers.put("message/rfc822", text);
        nutchParsers.put("text/css", text);
        nutchParsers.put("text/html", html);
        nutchParsers.put("text/plain", text);
        nutchParsers.put("text/richtext", msword);
        nutchParsers.put("text/rtf", msword);
        nutchParsers.put("text/sgml", html);
        nutchParsers.put("text/tab-separated-values", msexcel);
        nutchParsers.put("text/vnd.wap.wml", text);
        nutchParsers.put("text/vnd.wap.wmlscript", text);
        nutchParsers.put("text/xml", text);  // html, rss
        nutchParsers.put("text/x-setext", text);
    }

    public Analyzer getAnalyzer() {
        return analyzer;
    }
    
    /** Convenience method to subclasses to change the analyzer used for particular fields */
    protected void addAnalyzer(String fieldName, Analyzer analyzer) {
        this.analyzer.addAnalyzer(fieldName,analyzer);
    }
    
    /**
     * Returns a query parser to parse queries to this index.  Returns a new instance of QueryParser each time,
     * as these objects are lightweight and not thread-safe.
     * 
     * @return a query parser to parse queries to this index.
     */
    public QueryParser getQueryParser() {
        MultiFieldQueryParser searchParser = new MultiFieldQueryParser(Version.LUCENE_30,SEARCH_FIELDS, getAnalyzer());
        searchParser.setDefaultOperator(QueryParser.AND_OPERATOR);
        return searchParser;
    }
    
    public void setKeyRing(DOKeyRing keyring) {
        this.keyring = keyring;
    }

    @Override
	public String getSortFieldName(String field) {
		return "sort_" + field;
	}
    
    public Document documentOfStorageProxy(StorageProxy objStore) throws DOException, IOException {
        Document doc = new Document();
        doc.add(new Field("indexVersion",this.getClass().getName() + " " + VERSION,Field.Store.NO,Field.Index.ANALYZED));
        
        doc.add(new Field("id", objStore.getObjectID(), Field.Store.YES, Field.Index.NOT_ANALYZED));

        doc.add(new Field("repoid", objStore.getRepoID(), Field.Store.YES,
                Field.Index.NOT_ANALYZED));

        // store the object attributes in the index
        HeaderSet attributes = objStore.getAttributes(null);
        for(Iterator iter=attributes.iterator(); iter.hasNext(); ) {
            HeaderItem attribute = (HeaderItem)iter.next();
            doc.add(new Field("objatt_"+attribute.getName(),
                    attribute.getValue(),
                    Field.Store.YES, Field.Index.ANALYZED));
            doc.add(new Field(getSortFieldName("objatt_"+attribute.getName()),
            		attribute.getValue(),
            		Field.Store.NO, Field.Index.NOT_ANALYZED));
        }

        doc.add(new Field("objcreated", 
                DateTools.dateToString(new java.util.Date(attributes.getLongHeader(DOConstants.DATE_CREATED_ATTRIBUTE, 0)),
                        DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.NOT_ANALYZED));
        doc.add(new Field("objmodified", 
                DateTools.dateToString(new java.util.Date(attributes.getLongHeader(DOConstants.DATE_MODIFIED_ATTRIBUTE, 0)),
                        DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.NOT_ANALYZED));

        for(Enumeration en=objStore.listDataElements(); en.hasMoreElements(); ) {
            String elementID = (String)en.nextElement();
            addFieldForElement(doc, objStore.getObjectID(), objStore, elementID);
        }

        // if the internal.rights element exists, add a compiled-down form to the index
        if(objStore.doesDataElementExist(DOConstants.RIGHTS_ELEMENT_ID)) {
            net.cnri.apps.doserver.Authorizer authorizer = serverMain.getAuthorizer();
            String[] readers = authorizer.getExplicitAccepts(objStore, DOConstants.GET_DATA_OP_ID);
            StringBuffer readerStr = new StringBuffer();
            for(int i=0; readers!=null && i<readers.length; i++) {
                if(i!=0) readerStr.append('\n');
                readerStr.append(readers[i].replaceAll("\n", "\\n"));
            }
            doc.add(new Field("readers", readerStr.toString(), Field.Store.YES, Field.Index.ANALYZED));
            String[] writers = authorizer.getExplicitAccepts(objStore, DOConstants.STORE_DATA_OP_ID);
            StringBuffer writerStr = new StringBuffer();
            for(int i=0; writers!=null && i<writers.length; i++) {
                if(i!=0) writerStr.append('\n');
                writerStr.append(writers[i].replaceAll("\n", "\\n"));
            }
            doc.add(new Field("writers", writerStr.toString(), Field.Store.YES, Field.Index.ANALYZED));
        }


        //if(DEBUG) System.err.println("storing "+objID+" in index as doc:"+doc);
        doc.add(new Field("dtindexed",
                DateTools.dateToString(new Date(),
                        DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.NOT_ANALYZED));
        return doc;
    }
    
    /** Index the given data element and add the contents to the given lucene document. */
    private void addFieldForElement(Document doc, String objectID, StorageProxy objStore, String elementID) {
        if(DEBUG) System.err.println("  element: "+elementID);
        InputStream in = null;
        try {
            String defaultMimeType = null;
            if(elementID.equals(DOConstants.RIGHTS_ELEMENT_ID)) {
                defaultMimeType = "text/plain";
            }

            final String escapedElementID = escapeElementID(elementID);
            final boolean escapedSame = elementID.equals(escapedElementID);

            HeaderSet attributes = objStore.getElementAttributes(elementID, null);
            String mimeType = attributes.getStringHeader(DOConstants.MIME_TYPE_ATTRIBUTE, null);
            String title = attributes.getStringHeader(DOConstants.TITLE_ATTRIBUTE, null);
            String elementText = null;
            boolean encrypted = attributes.getStringHeader(DOKeyRing.KEY_ID_ATTRIBUTE, null)!=null;
            if(encrypted) {
                doc.add(new Field("has_enc_el", "1", Field.Store.YES, Field.Index.NOT_ANALYZED));
            }
            Metadata md = new Metadata();
            Parser parser = null;
            DOKeyRing keyring = this.keyring; // maintain thread safety
            if(mimeType!=null && (!encrypted || keyring!=null)) {
                if(mimeType.indexOf(';')>=0) {
                    mimeType = mimeType.substring(0, mimeType.indexOf(';')).trim();
                }
                md.set(Metadata.CONTENT_TYPE, mimeType);
                parser = (Parser)nutchParsers.get(mimeType.toLowerCase());

                // if we have a parser, retrieve the content
                if(parser!=null) {
                    in = objStore.getDataElement(elementID);

                    // read the contents of the element
                    ByteArrayOutputStream bout = new ByteArrayOutputStream();
                    byte buf[] = new byte[2048];
                    int r;
                    int numRead = 0;
                    while((r=in.read(buf))>=0) {
                        bout.write(buf, 0, r);
                        numRead += r;
                    }
                    try { in.close(); } catch (Throwable t) {}
                    md.set(Metadata.CONTENT_LENGTH, String.valueOf(numRead));

                    byte contentBytes[] = bout.toByteArray();
                    Content elementContent = new Content(elementID, "hdl:"+objectID,
                            contentBytes, mimeType, md, nutchConfig);
                    try {
                        Parse parse = parser.getParse(elementContent);
                        ParseData parseData = parse.getData();
                        Metadata parseMD = parseData.getContentMeta();
                        elementText = parse.getText();

                        if(title==null) title = parseData.getTitle();
                        String mdNames[] = parseMD.names();
                        for(int i=0; mdNames!=null && i<mdNames.length; i++) {
                            doc.add(new Field("elpatt_"+escapedElementID+"_"+mdNames[i], 
                                    parseMD.get(mdNames[i]),
                                    Field.Store.YES, Field.Index.NOT_ANALYZED));
                            if(!escapedSame) {
                                doc.add(new Field("elpatt_"+elementID+"_"+mdNames[i], 
                                        parseMD.get(mdNames[i]),
                                        Field.Store.NO, Field.Index.NOT_ANALYZED));
                            }
                        }
                        Metadata parseMD2 = parseData.getParseMeta();
                        mdNames = parseMD2.names();
                        for(int i=0; mdNames!=null && i<mdNames.length; i++) {
                            doc.add(new Field("elmpatt_"+escapedElementID+"_"+mdNames[i], 
                                    parseMD2.get(mdNames[i]),
                                    Field.Store.YES, Field.Index.NOT_ANALYZED));
                            if(!escapedSame) {
                                doc.add(new Field("elmpatt_"+elementID+"_"+mdNames[i], 
                                        parseMD2.get(mdNames[i]),
                                        Field.Store.NO, Field.Index.NOT_ANALYZED));
                            }
                        }
                    } catch (RuntimeException e) {
                        System.err.println("runtime parse error: "+e);
                    } catch (Exception e) {
                        System.err.println("parse error: "+e);
                    }
                }
            }

            String filename = attributes.getStringHeader(DOConstants.FILE_NAME_ATTRIBUTE, null);
            if((title==null || title.length()<=0) && filename!=null && filename.trim().length()>0) {
                title = filename;
            }

            if((title==null || title.length()<=0) && elementText!=null) {
                // try to pick out a decent title
                title = elementText;
                if(title.length()>200) {
                    title = title.substring(0, 200).trim();
                    int nlIdx = title.indexOf('\n');
                    if(nlIdx>0) title = title.substring(0, nlIdx).trim();
                } else {
                    title = elementText.trim();
                }
            }

            if(DEBUG) {
                String textToPrint = null;
                if(elementText!=null) {
                    textToPrint = elementText.trim();
                    if(textToPrint.length()>100)
                        textToPrint = textToPrint.substring(0, 100).trim();
                }
                System.err.println("  title: "+title);
                System.err.println("  mimetype: "+mimeType);
                System.err.println("  text: "+textToPrint);
            }

//            if(elementText==null || elementText.trim().length()<=0) {
//                // the parser failed
//                System.err.println("Warning: no text parsed from element "+elementID+" in object "+objectID);
//            }

            if(title!=null) {
                doc.add(new Field("eltitle_"+(elementID), title,
                        Field.Store.YES, Field.Index.ANALYZED));
            }
            if(mimeType!=null) {
                doc.add(new Field("eltype_"+(elementID), mimeType,
                        Field.Store.YES, Field.Index.NOT_ANALYZED));
            }
            if(elementText!=null) {
                doc.add(new Field("eltext_"+(elementID), elementText,
                        Field.Store.NO, Field.Index.ANALYZED));
            }
            for(Iterator iter=attributes.iterator(); iter.hasNext(); ) {
                HeaderItem attribute = (HeaderItem)iter.next();
                doc.add(new Field("elatt_"+escapedElementID+"_"+attribute.getName(), 
                        attribute.getValue(),
                        Field.Store.YES, Field.Index.ANALYZED));
                doc.add(new Field(getSortFieldName("elatt_"+escapedElementID+"_"+attribute.getName()),
                		attribute.getValue(),
                		Field.Store.NO, Field.Index.NOT_ANALYZED));
                if(!escapedSame) {
                    doc.add(new Field("elatt_"+elementID+"_"+attribute.getName(), 
                            attribute.getValue(),
                            Field.Store.NO, Field.Index.ANALYZED));
                    doc.add(new Field(getSortFieldName("elatt_"+elementID+"_"+attribute.getName()), 
                            attribute.getValue(),
                            Field.Store.NO, Field.Index.NOT_ANALYZED));
                }
            }
        } catch (Exception e) {
            System.err.println("error getting element "+elementID+" from object "+objectID+": "+e);
            if(DEBUG) {
                e.printStackTrace(System.err);
                System.err.flush();
            }
        } finally {
            try { if(in!=null) in.close(); } catch (Throwable t) {}
        }


    }

    public Document documentOfDigitalObject(DigitalObject obj) throws DOException, IOException {
        Document doc = new Document();
        try {
            doc.add(new Field("indexVersion",this.getClass().getName() + " " + VERSION,Field.Store.NO,Field.Index.ANALYZED));
            doc.add(new Field("id", obj.getID(), Field.Store.YES, Field.Index.NOT_ANALYZED));
            doc.add(new Field("repoid", obj.getRepository().getID(), Field.Store.YES,
                    Field.Index.NOT_ANALYZED));
            String[] elementIDs = obj.listDataElements();
            String attributes[] = obj.listAttributes();
            for(int i=0; attributes!=null && i<attributes.length; i++) {
                String attName = attributes[i];
                doc.add(new Field("objatt_"+attName,
                        obj.getAttribute(attName, ""),
                        Field.Store.YES, Field.Index.ANALYZED));
                doc.add(new Field(getSortFieldName("objatt_"+attName),
                		obj.getAttribute(attName, ""),
                		Field.Store.NO, Field.Index.NOT_ANALYZED));
            }
            doc.add(new Field("objcreated", 
                    DateTools.dateToString(obj.getDateCreated(),
                            DateTools.Resolution.MILLISECOND),
                            Field.Store.YES, Field.Index.NOT_ANALYZED));
            doc.add(new Field("objmodified", 
                    DateTools.dateToString(obj.getDateLastModified(),
                            DateTools.Resolution.MILLISECOND),
                            Field.Store.YES, Field.Index.NOT_ANALYZED));

            // add the different attributes to the document...
            for(int i=0; i<elementIDs.length; i++) {
                //if(!elementIDs[i].startsWith("content")) continue;
                addFieldForElement(doc, obj.getDataElement(elementIDs[i]));
            }
        } catch (DOException e) {
            System.err.println("Error indexing object "+obj.getID()+" in "+obj.getRepository()+": "+e);
            if(e.getErrorCode()!=DOException.PERMISSION_DENIED_ERROR) {
                throw e;
            }
            doc.add(new Field("idxerror", String.valueOf(e.getErrorCode()),
                    Field.Store.YES, Field.Index.NOT_ANALYZED));
        }

        //if(DEBUG) System.err.println("storing "+objID+" in index as doc:"+doc);
        doc.add(new Field("dtindexed",
                DateTools.dateToString(new Date(),
                        DateTools.Resolution.MILLISECOND),
                        Field.Store.YES, Field.Index.NOT_ANALYZED));

        return doc;
    }
    
    /** Initiate a retrieval of the given data element and add the contents to the
     * given lucene document.  Doesn't close the Reader that is passed as the Field
     * that is added to the Document. */
    private void addFieldForElement(Document doc, DataElement element) {
        if(DEBUG) System.err.println("  element: "+element);
        String elementID = element.getDataElementID();
        InputStream in = null;
        try {
            final String escapedElementID = escapeElementID(elementID);
            final boolean escapedSame = elementID.equals(escapedElementID);

            String mimeType = element.getAttribute(DOConstants.MIME_TYPE_ATTRIBUTE, null);
            String title = element.getAttribute(DOConstants.TITLE_ATTRIBUTE, null);
            String elementText = null;
            boolean encrypted = element.getAttribute(DOKeyRing.KEY_ID_ATTRIBUTE, null)!=null;
            if(encrypted) {
                doc.add(new Field("has_enc_el", "1", Field.Store.YES, Field.Index.NOT_ANALYZED));
            }
            Metadata md = new Metadata();
            Parser parser = null;
            DOKeyRing keyring = this.keyring; // maintain thread safety
            if(mimeType!=null && (!encrypted || keyring!=null)) {
                if(mimeType.indexOf(';')>=0) {
                    mimeType = mimeType.substring(0, mimeType.indexOf(';')).trim();
                }
                md.set(Metadata.CONTENT_TYPE, mimeType);
                parser = (Parser)nutchParsers.get(mimeType.toLowerCase());

                // if we have a parser, retrieve the content
                if(parser!=null) {
                    if(encrypted && keyring!=null) {
                        try {
                            in = keyring.decryptDataElement(element);
                        } catch (Exception e) {
                            System.err.println("Error decrypting element: "+e);
                            in = element.read();
                        }
                    } else {
                        in = element.read();
                    }

                    // read the contents of the element

                    ByteArrayOutputStream bout = new ByteArrayOutputStream();
                    byte buf[] = new byte[2048];
                    int r;
                    int numRead = 0;
                    while((r=in.read(buf))>=0) {
                        bout.write(buf, 0, r);
                        numRead += r;
                    }
                    try { in.close(); } catch (Throwable t) {}
                    md.set(Metadata.CONTENT_LENGTH, String.valueOf(numRead));

                    byte contentBytes[] = bout.toByteArray();
                    Content elementContent = 
                        new Content(elementID, "hdl:"+element.getDigitalObject().getID(),
                                contentBytes, mimeType, md, nutchConfig);

                    Parse parse = parser.getParse(elementContent);
                    ParseData parseData = parse.getData();
                    Metadata parseMD = parseData.getContentMeta();
                    elementText = parse.getText();

                    if(title==null) title = parseData.getTitle();
                    String mdNames[] = parseMD.names();
                    for(int i=0; mdNames!=null && i<mdNames.length; i++) {
                        doc.add(new Field("elpatt_"+escapedElementID+"_"+mdNames[i], 
                                parseMD.get(mdNames[i]),
                                Field.Store.YES, Field.Index.NOT_ANALYZED));
                        if(!escapedSame) {
                            doc.add(new Field("elpatt_"+elementID+"_"+mdNames[i], 
                                    parseMD.get(mdNames[i]),
                                    Field.Store.NO, Field.Index.NOT_ANALYZED));
                        }
                    }
                    Metadata parseMD2 = parseData.getParseMeta();
                    mdNames = parseMD2.names();
                    for(int i=0; mdNames!=null && i<mdNames.length; i++) {
                        doc.add(new Field("elmpatt_"+escapedElementID+"_"+mdNames[i], 
                                parseMD2.get(mdNames[i]),
                                Field.Store.YES, Field.Index.NOT_ANALYZED));
                        if(!escapedSame) {
                            doc.add(new Field("elmpatt_"+elementID+"_"+mdNames[i], 
                                    parseMD2.get(mdNames[i]),
                                    Field.Store.NO, Field.Index.NOT_ANALYZED));
                        }
                    }
                }
            }

            if((title==null || title.length()<=0) && elementText!=null) {
                // try to pick out a decent title
                title = elementText;
                if(title.length()>200) {
                    title = title.substring(0, 200).trim();
                    int nlIdx = title.indexOf('\n');
                    if(nlIdx>0) title = title.substring(0, nlIdx).trim();
                } else {
                    title = elementText.trim();
                }
            }

            if(DEBUG) {
                String textToPrint = null;
                if(elementText!=null) {
                    textToPrint = elementText.trim();
                    if(textToPrint.length()>100)
                        textToPrint = textToPrint.substring(0, 100).trim();
                }
                System.err.println("  title: "+title);
                System.err.println("  mimetype: "+mimeType);
                System.err.println("  text: "+textToPrint);

//                if(elementText==null || elementText.trim().length()<=0) {
//                    // the parser failed
//                    System.err.println("Warning: no text parsed from element "+element);
//                }
            }

            if(title!=null) {
                doc.add(new Field("eltitle_"+(elementID), title,
                        Field.Store.YES, Field.Index.ANALYZED));
            }
            if(mimeType!=null) {
                doc.add(new Field("eltype_"+(elementID), mimeType,
                        Field.Store.YES, Field.Index.NOT_ANALYZED));
            }
            if(elementText!=null) {
                doc.add(new Field("eltext_"+(elementID), elementText,
                        Field.Store.NO, Field.Index.ANALYZED));
            }
            String attributes[] = element.listAttributes();
            for(int i=0; attributes!=null && i<attributes.length; i++) {
                String attName = attributes[i];
                doc.add(new Field("elatt_"+escapedElementID+"_"+attName, 
                        element.getAttribute(attName, ""),
                        Field.Store.YES, Field.Index.ANALYZED));
                doc.add(new Field(getSortFieldName("elatt_"+escapedElementID+"_"+attName),
                		element.getAttribute(attName, ""),
                		Field.Store.NO, Field.Index.NOT_ANALYZED));
                if(!escapedSame) {
                    doc.add(new Field("elatt_"+elementID+"_"+attName, 
                            element.getAttribute(attName, ""),
                            Field.Store.NO, Field.Index.ANALYZED));
                    doc.add(new Field(getSortFieldName("elatt_"+elementID+"_"+attName),
                    		element.getAttribute(attName, ""),
                    		Field.Store.NO, Field.Index.NOT_ANALYZED));
                }
            }
        } catch (Exception e) {
            System.err.println("error getting element "+element+": "+e);
            if(DEBUG) {
                e.printStackTrace(System.err);
                System.err.flush();
            }
        } finally {
            try { if(in!=null) in.close(); } catch (Throwable t) {}
        }
    }


    
    private static String escapeElementID(String elementID) {
        return elementID.replace("%","%25").replace("_","%5F");
    }

    @Override
    public Query objectsNeedingReindexingQuery() {
        PhraseQuery versionQuery = new PhraseQuery();
        versionQuery.add(new Term("indexVersion",this.getClass().getName()));
        versionQuery.add(new Term("indexVersion",String.valueOf(VERSION)));
        BooleanQuery res = new BooleanQuery();
        res.add(new MatchAllDocsQuery(),Occur.MUST);
        res.add(versionQuery,Occur.MUST_NOT);
        return res;
    }
}
