/*************************************************************************\
    Copyright (c) 2012 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Robert Tupelo-Schneck <schneck@cnri.reston.va.us>
\*************************************************************************/
package net.cnri.apps.doserver.operators;

import java.io.IOException;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.IndexSearcher;

public class IndexSearcherManager {
    private IndexReader indexReader; // all access locked by readerLock
    private volatile LockTrackingIndexSearcherWrapper currentlyActiveIndexSearcherWrapper; // modifications locked by searcherLock
    private final Object readerLock = new Object();
    private final Object searcherLock = new Object();
    private volatile boolean reopenQueued;

    public IndexSearcherManager(IndexReader indexReader) {
        this.indexReader = indexReader;
        currentlyActiveIndexSearcherWrapper = new LockTrackingIndexSearcherWrapper();
    }

    public LockTrackingIndexSearcherWrapper getAndLockIndexSearcherWrapper() {
        reopenIfNecessary();
        synchronized(searcherLock) {
            currentlyActiveIndexSearcherWrapper.lock();
            return currentlyActiveIndexSearcherWrapper;
        }
    }
    
    private void reopenIfNecessary() {
        if(!reopenQueued) return;
        try {
            LockTrackingIndexSearcherWrapper oldIndexSearcherWrapper = null;
            synchronized(readerLock) {
                if(!reopenQueued) return;
                reopenQueued = false;
                IndexReader newReader = indexReader.reopen();
                if(newReader != indexReader) {
                    indexReader = newReader;
                    synchronized(searcherLock) {
                        oldIndexSearcherWrapper = currentlyActiveIndexSearcherWrapper;
                        currentlyActiveIndexSearcherWrapper = new LockTrackingIndexSearcherWrapper();
                    }
                }
            }
            if(oldIndexSearcherWrapper!=null) oldIndexSearcherWrapper.closeIfNeeded();
        }
        catch(Exception e) {
            System.err.println("Error reopening reader after update: "+e);
            e.printStackTrace(System.err);
        }
    }
    
    public void reopen() {
        reopenQueued = true;
    }

    public class LockTrackingIndexSearcherWrapper {
        private final AtomicInteger count = new AtomicInteger(0);
        private final IndexSearcher indexSearcher;

        private LockTrackingIndexSearcherWrapper() {
            indexSearcher = new IndexSearcher(indexReader);
        }
        
        public IndexSearcher getIndexSearcher() {
            return indexSearcher;
        }
        
        private void lock() { 
            count.incrementAndGet();
        }

        public void releaseAndCloseIfNeeded() { 
            if(count.decrementAndGet()==0 && this!=currentlyActiveIndexSearcherWrapper) close(); 
        }

        private void closeIfNeeded() { 
            if(count.get()==0 && this!=currentlyActiveIndexSearcherWrapper) close(); 
        }

        private void close() {
            try {
                IndexReader indexReader = this.indexSearcher.getIndexReader();
                this.indexSearcher.close();
                indexReader.close();
            }
            catch(IOException e) {}
        }

    }
}
