/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.apps.hdlstorage;

import net.handle.hdllib.*;
import net.handle.jdb.*;
import net.handle.server.*;

import java.io.*;
import java.util.*;

/** HDLCalculatedStorage is a HandleStorage module that can return handle values
  * for handles that don't exist in the handle database based on a template for
  * handles with a certain prefix.  For example, a query for the handle 123/12-555
  * will return the handle values for 123/12-555 if it exists, but if it doesn't
  * exist the values for 123/12 will be returned (if it exists).  This allows
  * us to eliminate the need to formally create handles for every digital object
  * that has the same handle values (ie handles that point to the same repository). 
  */
public class HDLCalculatedStorage 
  implements HandleStorage
{
  private File serverDir;
  private DBHash db;
  private DBHash naDB;
  private DBTransactionLog txnLog;

  private final String WRITE_LOCK = "WRITE_LOCK";
  private boolean readOnly = false;
  
  private final static byte[] BLANK_BYTES = new byte[0];
  private final static HandleValue[] BLANK_VALUE = new HandleValue[0];

  public HDLCalculatedStorage()
    throws Exception
  {
  }

  /** Initialize the JDB handle database with the given configuration.
      This currently doesn't do anything for the JDB database.
  */
  public void init(net.cnri.util.StreamTable config) {
    this.serverDir = new File(config.getStr("server_dir", "."));
                              
    try {
      db = new DBHash(new File(serverDir, "handles.jdb"), 5000, 1000);
      naDB = new DBHash(new File(serverDir, "nas.jdb"), 1000, 500);
    } catch (Throwable t) {
      System.err.println("Error initializing storage: "+t);
      t.printStackTrace(System.err);
      throw new RuntimeException("handle storage initialization error: "+t);
    }
  }
  
  /*********************************************************************
   * Returns true if this server is responsible for the given naming
   * authority.
   *********************************************************************/
  public final boolean haveNA(byte authHandle[])
    throws HandleException
  {
    try {
      return naDB.getValue(Util.upperCase(authHandle))!=null;
    } catch (Exception e) {
      throw new HandleException(HandleException.INTERNAL_ERROR,
                                "Error accessing NA data");
    }
  }
  
  /*********************************************************************
   * Sets a flag indicating whether or not this server is responsible
   * for the given naming authority.  
   *********************************************************************/
  public final void setHaveNA(byte authHandle[], boolean flag)
    throws HandleException
  {
    synchronized(WRITE_LOCK){
      // check if write operations are allowed.
      if(readOnly){
        throw new HandleException(HandleException.STORAGE_RDONLY);
      }
      
      try {
        byte[] handle = Util.upperCase(authHandle);
        if(flag){
          naDB.setValue(handle, BLANK_BYTES);
        } else {
          naDB.deleteValue(handle);
        }
      } catch(HandleException e1) {
        throw e1;
      } catch (Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  "Error accessing NA data");
      }
    } //end synchronized
  }


  /*********************************************************************
   * Return the pre-packaged values of the given handle that are either
   * in the indexList or the typeList.  This method should return any
   * values of type ALIAS or REDIRECT, even if they were not requested.
   *********************************************************************/
  public final byte[][] getRawHandleValues(byte inHandle[], int indexList[],
                                           byte typeList[][])
    throws HandleException
  {

    byte[] handle = inHandle;
    
    byte value[];
    try {
      value = db.getValue(handle);
    } catch (Exception e) {
      throw new HandleException(HandleException.INTERNAL_ERROR,
                                "Error retrieving handle");
    }
    if(value==null) {
      // if there is no overridden handle value, return a calculated value
      boolean gotSlash = false;
      for(int i=0; i<inHandle.length; i++) {
        if(!gotSlash) {
          gotSlash = inHandle[i]=='/';
        } else {
          if(inHandle[i]=='-') {
            byte baseHandle[] = new byte[i];
            System.arraycopy(inHandle, 0, baseHandle, 0, i);
            return getRawHandleValues(baseHandle, indexList, typeList);
          }
        }
      }
      return null;
    }
    
    int clumpLen;
    int bufPos = 0;
    boolean allValues =
       (indexList==null || indexList.length==0) &&
       (typeList==null  || typeList.length==0);
    int numValues = Encoder.readInt( value, bufPos);
    bufPos += Encoder.INT_SIZE;
    int origBufPos = bufPos;
    
    // figure out the number of records matching this request...
    int matches = 0;
    byte clumpType[];
    int clumpIndex;
    if(allValues) {
      matches = numValues;
    } else {
      for( int i=0; i < numValues; i++) {
        clumpLen = Encoder.readInt( value, bufPos);
        bufPos+=Encoder.INT_SIZE;
        
        clumpType = Encoder.getHandleValueType( value, bufPos);
        clumpIndex = Encoder.getHandleValueIndex( value, bufPos);
        
        if( Util.isParentTypeInArray( typeList, clumpType) ||
            Util.isInArray( indexList, clumpIndex))
           matches++;
        
        bufPos+=clumpLen;
      }
    }
	
    // populate and return an array with the matched records...
    byte clumps[][] = new byte[matches][];
    int clumpNum = 0;
    bufPos = origBufPos;
    for(int i=0; i<numValues; i++) {
      clumpLen = Encoder.readInt( value, bufPos);
      bufPos+=Encoder.INT_SIZE;
      
      clumpType = Encoder.getHandleValueType( value, bufPos);
      clumpIndex = Encoder.getHandleValueIndex( value, bufPos);
      
      if( allValues ||
          Util.isParentTypeInArray( typeList, clumpType) ||
          Util.isInArray( indexList, clumpIndex)) {
        clumps[clumpNum] = new byte[clumpLen];
        System.arraycopy( value, bufPos, clumps[clumpNum], 0, clumpLen);
        clumpNum++;
      }
      
      bufPos+=clumpLen;
    }
    return clumps;
  }
  


  
  /*********************************************************************
   * Creates the specified handle in the "database" with the specified
   * initial values
   *********************************************************************/
  public final void createHandle(byte inHandle[], HandleValue values[])
    throws HandleException
  {
    synchronized(WRITE_LOCK) {
      // check if write operations are allowed.
      if(readOnly){
        throw new HandleException(HandleException.STORAGE_RDONLY);
      }

      byte[] handle = inHandle;
      
      byte existingValue[] = null;
      try {
        existingValue = db.getValue(handle);
      } catch(Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  "Error checking for existing handle");
      }
      if (existingValue!=null)
        throw new HandleException(HandleException.HANDLE_ALREADY_EXISTS,
                                  "Handle already exists");
      
      // add the size of all of the clumps as well as a 'time-received' field
      // and 'value-length' field for each value
      int totalSize = Encoder.INT_SIZE;
      for( int i=0; i < values.length; i++) {
        totalSize += ( Encoder.calcStorageSize(values[i]) + Encoder.INT_SIZE);
      }
      
      // write all of the clumps to a buffer
      byte data[] = new byte[totalSize];
      int offst = 0;
      offst += Encoder.writeInt( data, offst, values.length);
      for( int i=0; i < values.length; i++) {
        int clumpLen = Encoder.encodeHandleValue( data, offst+Encoder.INT_SIZE,
                                                  values[i]);
        offst += Encoder.writeInt(data, offst, clumpLen);
        offst += clumpLen;
      }
      
      try {
        db.setValue(handle,data);
      } catch(HandleException e1) {
        throw e1;
      } catch (Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  "Error creating handle");
      }
    } //end synchronized
  }
  
  /*********************************************************************
   * Delete the specified handle in the database.
   *********************************************************************/
  public final boolean deleteHandle(byte inHandle[])
    throws HandleException
  {
    synchronized(WRITE_LOCK){
      //check if write operations are allowed.
      if(readOnly){
        throw new HandleException(HandleException.STORAGE_RDONLY);
      }

      byte[] handle = inHandle;
      
      boolean deleted;
      try {
        deleted = db.deleteValue(handle);
      } catch(HandleException e1) {
        throw e1;
      } catch (Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  "Error deleting handle");
      }
      return deleted;
    }//end synchronized
  }
  
  /*********************************************************************
   * Replace the current values for the given handle with new values.
   *********************************************************************/
  public final void updateValue(byte inHandle[], HandleValue values[])
    throws HandleException
  {
    synchronized(WRITE_LOCK) {
      // check if write operations are allowed.
      if(readOnly) {
        throw new HandleException(HandleException.STORAGE_RDONLY);
      }

      byte[] handle = inHandle;
      
      byte existingValue[] = null;
      try {
        existingValue = db.getValue(handle);
      } catch(Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  "Error checking for existing handle");
      }
      if (existingValue==null)
        throw new HandleException(HandleException.HANDLE_DOES_NOT_EXIST,
                                  "Cannot modify non-existent handle");
      
      // add the size of all of the clumps as well as a 'time-received' field
      // and 'value-length' field for each value
      int totalSize = Encoder.INT_SIZE;
      for( int i=0; i < values.length; i++) {
        totalSize += ( Encoder.calcStorageSize(values[i]) + Encoder.INT_SIZE);
      }
      
      // write all of the clumps to a buffer
      byte data[] = new byte[totalSize];
      int offst = 0;
      offst += Encoder.writeInt( data, offst, values.length);
      for( int i=0; i < values.length; i++) {
        int clumpLen = Encoder.encodeHandleValue( data, offst+Encoder.INT_SIZE,
                                                  values[i]);
        offst += Encoder.writeInt(data, offst, clumpLen);
        offst += clumpLen;
      }
      
      try {
        db.setValue(handle, data);
      } catch(HandleException e1) {
        throw e1;
      } catch (Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  "Error updating handle");
      }
    } //end synchronized
  }
  
  /*********************************************************************
   * Scan the database, calling a method in the specified callback for 
   * every handle in the database.
   *********************************************************************/
  public final void scanHandles(ScanCallback callback)
    throws HandleException
  {
    for(java.util.Enumeration recs=db.getEnumerator(); recs.hasMoreElements(); ) {
      byte[][] record = (byte[][])recs.nextElement();
      callback.scanHandle(record[0]);
    }
  }
  
  /*********************************************************************
   * Scan the NA database, calling a method in the specified callback for 
   * every naming authority handle in the database.
   *********************************************************************/
  public void scanNAs(ScanCallback callback)
    throws HandleException
  {
    for(java.util.Enumeration recs=naDB.getEnumerator(); recs.hasMoreElements(); ) {
      byte[][] record = (byte[][])recs.nextElement();
      callback.scanHandle(record[0]);
    }
  }
  
  /*********************************************************************
   * Scan the database for handles with the given naming authority
   * and return an Enumeration of byte arrays with each byte array
   * being a handle.  <i>naHdl</i> is the naming authority handle
   * for the naming authority that you want to list the handles for.
   *********************************************************************/
  public final Enumeration getHandlesForNA(byte naHdl[])
    throws HandleException
  {
    if(!haveNA(naHdl)) {
      throw new HandleException(HandleException.INVALID_VALUE,
                                "The requested naming authority doesn't live here");
    }
    
    return new HdlsForNAEnum(db.getEnumerator(), null); //Util.getIDPart(naHdl));
  }
  
  /**********************************************************************
   * This is an class that scans all of the records in a database and
   * forwards only the handles that start with the given prefix.
   **********************************************************************/
  private final class HdlsForNAEnum
    implements Enumeration
  {
    private byte prefix[];
    private byte nextHdl[] = null;
    private Enumeration dbEnum;
    
    HdlsForNAEnum(Enumeration dbEnum, byte na[]) {
      this.prefix = Util.encodeString(Util.decodeString(na)+'/');
      this.dbEnum = dbEnum;
      seekNextValue();
    }
    
    private final void seekNextValue() {
      nextHdl = null;
      byte tmpRecord[][];
      while(dbEnum.hasMoreElements()) {
        tmpRecord = (byte[][])dbEnum.nextElement();
        if(Util.startsWithCI(tmpRecord[0], prefix)) {
          nextHdl = tmpRecord[0];
          return;
        }
      }
    }
    
    public final boolean hasMoreElements() {
      return nextHdl!=null;
    }
    
    public final Object nextElement() {
      byte thisHdl[] = nextHdl;
      seekNextValue();
      return thisHdl;
    }
  }
  
  /*********************************************************************
   * Remove all of the records from the database.
   ********************************************************************/
  public final void deleteAllRecords()
    throws HandleException
  {
    synchronized(WRITE_LOCK) {
      // check if write operations are allowed.
      if(readOnly){
        throw new HandleException(HandleException.STORAGE_RDONLY);
      }
      
      try {
        db.deleteAllRecords();
        naDB.deleteAllRecords();
      } catch(HandleException e1) {
        throw e1;
      } catch (Exception e) {
        throw new HandleException(HandleException.INTERNAL_ERROR,
                                  String.valueOf(e));
      }
    } //end synchronized
  }
  

  /*********************************************************************
   * Copy the current database to a backup file, and restart the transaction log.
   * Returns true if the checkpoint process was *started* (this doesn't mean
   * it was successful).
   *********************************************************************/
  public final void checkpointDatabase()
    throws HandleException
  {
    throw new HandleException(HandleException.INVALID_VALUE, "Transaction logging not enabled");
  }
  
  /*********************************************************************
   * Close the database and clean up
   *********************************************************************/
  public final void shutdown() {
    try {
      db.close();
    } catch (Throwable e) {}
    try {
      naDB.close();
    } catch (Throwable e) {}
  }

  public void finalize() {
    shutdown();
  }
  
}
