/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.dobj;

import java.net.*;
import java.util.*;

import java.security.*;
import java.security.cert.*;
import javax.net.ssl.*;

import net.handle.hdllib.Resolver;

import org.bouncycastle.jce.X509Principal;
import org.bouncycastle.x509.X509V3CertificateGenerator;
import org.bouncycastle.asn1.x509.X509Name;


/** This object is used to provide a private key for authentication to the other side
 * of a secure socket connection.
 */
public class DOSSLKeyManager
  implements X509KeyManager
{
  private static final String DOP_AUTH_ALIAS = "DOP_AUTH_ALIAS";
  private X509Certificate myCert = null;
  private PKAuthentication myAuth = null;
  private PrivateKey privKey = null;
  private PublicKey publicKey = null;
  private X509V3CertificateGenerator v3CertGen = null;
  private long nextExpiration = Long.MIN_VALUE;
  private boolean isStandardCert = false;
  private KeyManager standardKM;



  public DOSSLKeyManager(PKAuthentication auth) 
    throws Exception
  {
      this(auth, null);
  }
  

  public DOSSLKeyManager(PKAuthentication auth, PublicKey pubKey)
  throws Exception
  {
    this.myAuth = auth;
    
    privKey = myAuth.getPrivateKey();
    v3CertGen = new X509V3CertificateGenerator();
    
    if(!auth.getID().equals(DOConstants.ANONYMOUS_ID)) {
    	
      String privKeyAlg = privKey.getAlgorithm();
      // generate a signature that will be used to verify a match with the public key
      Signature signer = Signature.getInstance("SHA1with"+privKeyAlg);
      signer.initSign(privKey);
      byte sigTestData[] = "hello, how are you?".getBytes();
      signer.update(sigTestData);
      byte testSig[] = signer.sign();
        
      PublicKey[] pubKeys;
      if(pubKey==null) {
          Resolver temp = DOClient.getResolver();
          String tempID = auth.getID();
          pubKeys = temp.resolvePublicKeys(tempID);
      }
      else {
          pubKeys = new PublicKey[] { pubKey };
          pubKey = null;
      }
          
      for(int i=0; pubKeys!=null && i<pubKeys.length; i++) {
        if(pubKeys[i]==null) continue;
        if(!pubKeys[i].getAlgorithm().equals(privKeyAlg)) continue;
        
        Signature verifier = Signature.getInstance("SHA1with"+privKeyAlg);
        verifier.initVerify(pubKeys[i]);
        verifier.update(sigTestData);
        if(verifier.verify(testSig)) {
          pubKey = pubKeys[i];
        }
      }
      
      if(pubKey==null) {
        throw new DOException(DOException.CRYPTO_ERROR, "Public keys resolved for "+
                              myAuth.getID()+" do not match supplied private key");
      }
    }
    this.publicKey = pubKey;
    //cert.verify(myPubKey);
  }
  
  
  
  private synchronized X509Certificate getMyCert() {
    long now = System.currentTimeMillis();
    if(now < nextExpiration) return myCert;
    
    // generate my own certificate and compile list of group certificates
    
    // create the certificate
    Hashtable myAttributes = new Hashtable();
    //myAttributes.put("o", DERObjectIdentifier.getInstance(auth.getID()));
    myAttributes.put(X509Name.O, myAuth.getID());
    //myAttributes.put(X509Name.OU, "DO Certificate");
    
    Hashtable subjectAttributes = new Hashtable();
    subjectAttributes.put(X509Name.O, myAuth.getID());
    //subjectAttributes.put("ou", "DO Certificate");
    
    v3CertGen.setSerialNumber(java.math.BigInteger.valueOf(System.currentTimeMillis()));
    v3CertGen.setIssuerDN(new X509Principal(myAttributes));
    v3CertGen.setNotBefore(new Date(System.currentTimeMillis() - 1000L * 60 * 60 * 24 * 30));
    v3CertGen.setNotAfter(new Date(System.currentTimeMillis() + (1000L * 60 * 60 * 24 * 30)));
    v3CertGen.setSubjectDN(new X509Principal(subjectAttributes));
    v3CertGen.setPublicKey(publicKey);
    v3CertGen.setSignatureAlgorithm("SHA1with"+privKey.getAlgorithm());
    
    try {
      myCert = v3CertGen.generateX509Certificate(privKey);
      myCert.checkValidity(new Date());
    } catch (Exception e) {
      System.err.println("Error creating self-signed certificate: "+e);
      e.printStackTrace(System.err);
    }
    
    nextExpiration = now + 864000 * 1000; // re-generate the cert every day
    return myCert;
  }
  
  public String chooseClientAlias(String keyTypes[], Principal issuers[], Socket socket) {
    if(getMyCert()==null) return null;
    return DOP_AUTH_ALIAS;
  }


  public String chooseServerAlias(String keyType, Principal[] issuers, Socket socket) {
    if(getMyCert()==null) return null;
    return DOP_AUTH_ALIAS;
  }
  
  public String[] getServerAliases(String keyType, Principal[] issuers) {
    return myAuth!=null ? new String[] { DOP_AUTH_ALIAS } : new String[0];
  }
  
  public X509Certificate[] getCertificateChain(String alias) {
    if(getMyCert()!=null) {
      return new X509Certificate[] { getMyCert() };
    } else {
      return new X509Certificate[0];
    }
  }
  
  public String[] getClientAliases(String keyType, Principal issuers[]) {
    return myAuth!=null ? new String[] { DOP_AUTH_ALIAS } : new String[0];
  }
  
  public PrivateKey getPrivateKey(String alias) {
    return myAuth!=null ? myAuth.getPrivateKey() : null;
  }
  
}

